from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal

import torch

from graph_pes.atomic_graph import (
    AtomicGraph,
    number_of_atoms,
    number_of_edges,
    number_of_neighbours,
    sum_over_neighbours,
)
from graph_pes.utils.misc import (
    is_being_documented,
    left_aligned_div,
    uniform_repr,
)

if TYPE_CHECKING or is_being_documented():
    NeighbourAggregationMode = Literal[
        "sum", "mean", "constant_fixed", "constant_learnable", "sqrt"
    ]
else:
    NeighbourAggregationMode = str


class NeighbourAggregation(ABC, torch.nn.Module):
    r"""
    An abstract base class for aggregating values over neighbours:

    .. math::

        X_i^\prime = \text{Agg}_{j \in \mathcal{N}_i} \left[X_j\right]

    where :math:`\mathcal{N}_i` is the set of neighbours of atom :math:`i`,
    :math:`X` has shape ``(E, ...)``, :math:`X^\prime` has shape ``(N, ...)``
    and ``E`` and ``N`` are the number of edges and atoms in the graph,
    respectively.
    """

    @abstractmethod
    def forward(self, x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        """Aggregate x over neighbours."""

    def pre_fit(self, graphs: AtomicGraph) -> None:
        """
        Calculate any quantities that are dependent on the graph structure
        that should be fixed before prediction.

        Default implementation does nothing.

        Parameters
        ----------
        graphs
            A batch of graphs to pre-fit to.
        """

    # type hints for mypy etc.
    def __call__(self, x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        return super().__call__(x, graph)

    @staticmethod
    def parse(
        mode: Literal[
            "sum", "mean", "constant_fixed", "constant_learnable", "sqrt"
        ],
    ) -> NeighbourAggregation:
        """
        Evaluates the following map:

        .. list-table::
           :widths: 30 70
           :header-rows: 1

           * - Mode
             - Aggregation
           * - ``"sum"``
             - :class:`SumNeighbours() <SumNeighbours>`
           * - ``"mean"``
             - :class:`MeanNeighbours() <MeanNeighbours>`
           * - ``"constant_fixed"``
             - :class:`ScaledSumNeighbours(learnable=False) <ScaledSumNeighbours>`
           * - ``"constant_learnable"``
             - :class:`ScaledSumNeighbours(learnable=True) <ScaledSumNeighbours>`
           * - ``"sqrt"``
             - :class:`VariancePreservingSumNeighbours() <VariancePreservingSumNeighbours>`

        Parameters
        ----------
        mode
            The neighbour aggregation mode to parse.

        Returns
        -------
        NeighbourAggregation
            The parsed neighbour aggregation mode.
        """  # noqa: E501
        if mode == "sum":
            return SumNeighbours()
        elif mode == "mean":
            return MeanNeighbours()
        elif mode == "constant_fixed":
            return ScaledSumNeighbours(learnable=False)
        elif mode == "constant_learnable":
            return ScaledSumNeighbours(learnable=True)
        elif mode == "sqrt":
            return VariancePreservingSumNeighbours()
        else:
            raise ValueError(f"Unknown neighbour aggregation mode: {mode}")


class SumNeighbours(NeighbourAggregation):
    r"""
    Sum over neighbours:

    .. math::

        X_i^\prime = \sum_{j \in \mathcal{N}_i} X_j
    """

    def forward(self, x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        return sum_over_neighbours(x, graph)


class MeanNeighbours(NeighbourAggregation):
    r"""
    Take an average over neighbours:

    .. math::

        X_i^\prime = \frac{1}{|\mathcal{N}_i|} \sum_{j \in \mathcal{N}_i} X_j

    where :math:`|\mathcal{N}_i|` is the number of neighbours of atom :math:`i`
    (including the central atom).

    .. note::

        This aggregation can lead to un-physical discontinuities in the PES
        as neighbours enter or leave the radial cutoff.
    """

    def forward(self, x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        return left_aligned_div(
            sum_over_neighbours(x, graph),
            number_of_neighbours(graph, include_central_atom=True),
        )


class ScaledSumNeighbours(NeighbourAggregation):
    r"""
    Scale the sum over neighbours by a learnable or fixed constant,
    :math:`s`:

    .. math::

        X_i^\prime = \frac{1}{s} \sum_{j \in \mathcal{N}_i} X_j

    :math:`s` defaults to 1.0, but is set to the average number of neighbours
    of each atom in the training set passed to :meth:`pre_fit`.

    Parameters
    ----------
    learnable
        If ``True``, the scale is a learnable parameter. If ``False``, the
        scale is a fixed constant.
    """

    def __init__(self, learnable: bool = False):
        super().__init__()
        self.scale = torch.nn.Parameter(
            torch.tensor(1.0), requires_grad=learnable
        )

    def forward(self, x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        return sum_over_neighbours(x, graph) / self.scale

    def pre_fit(self, graphs: AtomicGraph) -> None:
        """
        Set the scale equal to the average number of neighbours in the
        training set.
        """
        avg_neighbours = number_of_edges(graphs) / number_of_atoms(graphs)
        self.scale.data = torch.tensor(avg_neighbours)

    def __repr__(self) -> str:
        return uniform_repr(
            self.__class__.__name__,
            scale=f"{self.scale.item():.3f}",
            learnable=self.scale.requires_grad,
        )


class VariancePreservingSumNeighbours(NeighbourAggregation):
    r"""
    Scale the sum over neighbours by the square root of the number of
    neighbours:

    .. math::

        X_i^\prime = \frac{1}{\sqrt{|\mathcal{N}_i|}} \sum_{j \in \mathcal{N}_i} X_j

    where :math:`|\mathcal{N}_i|` is the number of neighbours of atom :math:`i`
    (including the central atom).

    .. note::

        This aggregation can lead to un-physical discontinuities in the PES
        as neighbours enter or leave the radial cutoff.
    """  # noqa: E501

    def forward(self, x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
        return left_aligned_div(
            sum_over_neighbours(x, graph),
            torch.sqrt(number_of_neighbours(graph, include_central_atom=True)),
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Final, cast

import torch
import torch.fx
from e3nn import o3

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    index_over_neighbours,
    neighbour_distances,
    neighbour_vectors,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components.aggregation import (
    NeighbourAggregation,
    NeighbourAggregationMode,
)
from graph_pes.models.components.distances import (
    DistanceExpansion,
    PolynomialEnvelope,
    get_distance_expansion,
)
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.models.e3nn.mace_utils import (
    Contraction,
    ContractionConfig,
    UnflattenIrreps,
    parse_irreps,
)
from graph_pes.models.e3nn.utils import (
    LinearReadOut,
    NonLinearReadOut,
    ReadOut,
    SphericalHarmonics,
    as_irreps,
    build_limited_tensor_product,
    to_full_irreps,
)
from graph_pes.utils.nn import (
    MLP,
    AtomicOneHot,
    HaddamardProduct,
    MLPConfig,
    PerElementEmbedding,
    UniformModuleList,
)


class MACEInteraction(torch.nn.Module):
    """
    The MACE interaction block.

    Generates new node embeddings from the old node embeddings and the
    spherical harmonic expansion and mangitudes of the neighbour vectors.
    """

    def __init__(
        self,
        # input nodes
        irreps_in: list[o3.Irrep],
        nodes: NodeDescription,
        # input edges
        sph_harmonics: o3.Irreps,
        radial_basis_features: int,
        mlp: MLPConfig,
        # other
        aggregation: NeighbourAggregationMode,
        mix_attributes: bool,
    ):
        super().__init__()

        irreps_out = [ir for _, ir in sph_harmonics]

        features_in = as_irreps([(nodes.channels, ir) for ir in irreps_in])
        self.pre_linear = o3.Linear(
            features_in,
            features_in,
            internal_weights=True,
            shared_weights=True,
        )

        self.tp = build_limited_tensor_product(
            features_in,
            sph_harmonics,
            irreps_out,
        )
        mid_features = self.tp.irreps_out.simplify()
        assert all(ir in mid_features for ir in irreps_out)

        self.weight_generator = MLP.from_config(
            mlp,
            input_features=radial_basis_features,
            output_features=self.tp.weight_numel,
            bias=False,
        )

        features_out = as_irreps(
            [(nodes.channels, ir) for (_, ir) in sph_harmonics]
        )
        self.post_linear = o3.Linear(
            mid_features,
            features_out,
            internal_weights=True,
            shared_weights=True,
        )

        self.aggregator = NeighbourAggregation.parse(aggregation)

        if mix_attributes:
            self.attribute_mixer = o3.FullyConnectedTensorProduct(
                irreps_in1=features_out,
                irreps_in2=o3.Irreps(f"{nodes.attributes}x0e"),
                irreps_out=features_out,
            )
        else:
            self.attribute_mixer = None

        self.reshape = UnflattenIrreps(irreps_out, nodes.channels)

        # book-keeping
        self.irreps_in = features_in
        self.irreps_out = features_out

    def forward(
        self,
        node_features: torch.Tensor,
        node_attributes: torch.Tensor,
        sph_harmonics: torch.Tensor,
        radial_basis: torch.Tensor,
        graph: AtomicGraph,
    ) -> torch.Tensor:
        # pre-linear
        node_features = self.pre_linear(node_features)  # (N, a)

        # tensor product: mix node and edge features
        neighbour_features = index_over_neighbours(
            node_features, graph
        )  # (E, a)
        weights = self.weight_generator(radial_basis)  # (E, b)
        messages = self.tp(
            neighbour_features,
            sph_harmonics,
            weights,
        )  # (E, c)

        # aggregate
        total_message = self.aggregator(messages, graph)  # (N, c)

        # post-linear
        node_features = self.post_linear(total_message)  # (N, d)

        if self.attribute_mixer is not None:
            node_features = self.attribute_mixer(node_features, node_attributes)

        return self.reshape(node_features)  # (N, channels, d')

    # type hints for mypy
    def __call__(
        self,
        node_features: torch.Tensor,
        node_attributes: torch.Tensor,
        sph_harmonics: torch.Tensor,
        radial_basis: torch.Tensor,
        graph: AtomicGraph,
    ) -> torch.Tensor:
        return super().__call__(
            node_features,
            node_attributes,
            sph_harmonics,
            radial_basis,
            graph,
        )


@dataclass
class NodeDescription:
    channels: int
    attributes: int
    hidden_features: list[o3.Irrep]

    def hidden_irreps(self) -> o3.Irreps:
        return to_full_irreps(self.channels, self.hidden_features)


class MACELayer(torch.nn.Module):
    def __init__(
        self,
        irreps_in: list[o3.Irrep],
        nodes: NodeDescription,
        correlation: int,
        sph_harmonics: o3.Irreps,
        radial_basis_features: int,
        mlp: MLPConfig,
        use_sc: bool,
        aggregation: NeighbourAggregationMode,
        residual: bool,
        final_layer: bool,
    ):
        super().__init__()

        self.interaction = MACEInteraction(
            irreps_in=irreps_in,
            nodes=nodes,
            sph_harmonics=sph_harmonics,
            radial_basis_features=radial_basis_features,
            mlp=mlp,
            aggregation=aggregation,
            # only mix attributes in the interaction block
            # if we **aren't** using a residual connection
            mix_attributes=not residual,
        )
        actual_mid_features = [ir for _, ir in self.interaction.irreps_out]

        output_features = o3.Irreps(
            nodes.hidden_irreps()
            if not final_layer
            else o3.Irreps(f"{nodes.channels}x0e")
        )
        self.contractions = UniformModuleList(
            [
                Contraction(
                    config=ContractionConfig(
                        num_features=nodes.channels,
                        n_node_attributes=nodes.attributes,
                        irrep_s_in=actual_mid_features,
                        irrep_out=target_irrep,
                    ),
                    correlation=correlation,
                )
                for target_irrep in [o.ir for o in output_features]
            ]
        )

        if use_sc and residual:
            # links input features to output features via a tensor product
            self.residual_update = o3.FullyConnectedTensorProduct(
                irreps_in1=[(nodes.channels, ir) for ir in irreps_in],
                irreps_in2=o3.Irreps(f"{nodes.attributes}x0e"),
                irreps_out=output_features,
            )
        else:
            self.residual_update = None

        # update the hidden features from the interaction block
        # and target the output features
        self.post_linear = o3.Linear(
            output_features,
            output_features,
            internal_weights=True,
            shared_weights=True,
        )

        # book-keeping
        self.irreps_in = irreps_in
        self.irreps_out: o3.Irreps = output_features  # type: ignore

    def forward(
        self,
        node_features: torch.Tensor,
        node_attributes: torch.Tensor,
        sph_harmonics: torch.Tensor,
        radial_basis: torch.Tensor,
        graph: AtomicGraph,
    ) -> torch.Tensor:
        # A MACE layer operates on:
        # - node features with multiplicity M, e.g. M=16: 16x0e + 16x1o
        # - node attributes with multiplicity A e.g. A=5: 5x0e
        # - spherical harmonics up to l_max, e.g. l_max=2: 1x0e + 1x1o + 1x2e

        # interact
        internal_node_features = self.interaction(
            node_features,
            node_attributes,
            sph_harmonics,
            radial_basis,
            graph,
        )  # (N, M, irreps)

        # contract using the contractions directly
        contracted_features = torch.cat(
            [
                contraction(internal_node_features, node_attributes)
                for contraction in self.contractions
            ],
            dim=-1,
        )  # (N, irreps_out)

        # residual update
        if self.residual_update is not None:
            update = self.residual_update(
                node_features,
                node_attributes,
            )  # (N, irreps_out)
            contracted_features = contracted_features + update

        # linear update
        node_features = self.post_linear(contracted_features)  # (N, irreps_out)

        return node_features


class _BaseMACE(GraphPESModel):
    def __init__(
        self,
        # radial things
        cutoff: float,
        n_radial: int,
        radial_expansion: type[DistanceExpansion] | str,
        weights_mlp: MLPConfig,
        # node things
        nodes: NodeDescription,
        node_attribute_generator: Callable[[torch.Tensor], torch.Tensor],
        # message passing
        layers: int,
        l_max: int,
        correlation: int,
        neighbour_aggregation: NeighbourAggregationMode,
        use_self_connection: bool,
        # readout
        readout_width: int,
    ):
        super().__init__(
            cutoff=cutoff,
            implemented_properties=["local_energies"],
        )

        if o3.Irrep("0e") not in nodes.hidden_features:
            raise ValueError("MACE requires a `0e` hidden feature")

        # radial things
        sph_harmonics = cast(o3.Irreps, o3.Irreps.spherical_harmonics(l_max))
        self.spherical_harmonics = SphericalHarmonics(
            sph_harmonics,
            normalize=True,
            normalization="component",
        )
        if isinstance(radial_expansion, str):
            radial_expansion = get_distance_expansion(radial_expansion)
        self.radial_expansion = HaddamardProduct(
            radial_expansion(
                n_features=n_radial, cutoff=cutoff, trainable=True
            ),
            PolynomialEnvelope(cutoff=cutoff, p=5),
        )

        # node things
        self.node_attribute_generator = node_attribute_generator
        self.initial_node_embedding = PerElementEmbedding(nodes.channels)

        # message passing
        current_node_irreps = [o3.Irrep("0e")]
        self.layers: UniformModuleList[MACELayer] = UniformModuleList([])

        for i in range(layers):
            # only use residual skip after the first layer
            use_residual = i != 0
            final_layer = i == layers - 1
            layer = MACELayer(
                irreps_in=current_node_irreps,
                nodes=nodes,
                correlation=correlation,
                sph_harmonics=sph_harmonics,
                radial_basis_features=n_radial,
                mlp=weights_mlp,
                use_sc=use_self_connection,
                aggregation=neighbour_aggregation,
                residual=use_residual,
                final_layer=final_layer,
            )
            self.layers.append(layer)
            current_node_irreps = [ir for _, ir in layer.irreps_out]

        self.readouts: UniformModuleList[ReadOut] = UniformModuleList(
            [LinearReadOut(nodes.hidden_irreps()) for _ in range(layers - 1)]
            + [
                NonLinearReadOut(
                    self.layers[-1].irreps_out, hidden_dim=readout_width
                )
            ],
        )

        self.scaler = LocalEnergiesScaler()

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        # pre-compute some things
        vectors = neighbour_vectors(graph)
        sph_harmonics = self.spherical_harmonics(vectors)
        edge_features = self.radial_expansion(
            neighbour_distances(graph).view(-1, 1)
        )
        node_attributes = self.node_attribute_generator(graph.Z)

        # generate initial node features
        node_features = self.initial_node_embedding(graph.Z)

        # update node features through message passing layers
        per_atom_energies = []
        for layer, readout in zip(self.layers, self.readouts):
            node_features = layer(
                node_features,
                node_attributes,
                sph_harmonics,
                edge_features,
                graph,
            )
            per_atom_energies.append(readout(node_features))

        # sum up the per-atom energies
        local_energies = torch.sum(
            torch.stack(per_atom_energies), dim=0
        ).squeeze()

        # return scaled local energy predictions
        return {"local_energies": self.scaler(local_energies, graph)}


DEFAULT_MLP_CONFIG: Final[MLPConfig] = {
    "hidden_depth": 3,
    "hidden_features": 64,
    "activation": "SiLU",
}


class MACE(_BaseMACE):
    r"""
    The `MACE <https://arxiv.org/abs/2206.07697>`__ architecture.

    One-hot encodings of the atomic numbers are used to condition the
    ``TensorProduct`` update in the residual connection of the message passing
    layers, as well as the contractions in the message passing layers.

    Following the notation used in `ACEsuite/mace <https://github.com/ACEsuit/mace>`__,
    the first layer in this model is a ``RealAgnosticInteractionBlock``. Subsequent
    layers are then ``RealAgnosticResidualInteractionBlock``\ s

    Please cite the following if you use this model in your research:

    .. code-block:: bibtex

        @misc{Batatia2022MACE,
            title = {
                MACE: Higher Order Equivariant Message Passing
                Neural Networks for Fast and Accurate Force Fields
            },
            author = {
                Batatia, Ilyes and Kov{\'a}cs, D{\'a}vid P{\'e}ter and
                Simm, Gregor N. C. and Ortner, Christoph and Cs{\'a}nyi, G{\'a}bor
            },
            year = {2022},
            doi = {10.48550/arXiv.2206.07697},
        }

    Parameters
    ----------
    elements
        list of elements that this MACE model will be able to handle.
    cutoff
        radial cutoff (in Å) for the radial expansion (and message passing)
    n_radial
        number of bases to expand the radial distances into
    radial_expansion
        type of radial expansion to use. See :class:`~graph_pes.models.components.distances.DistanceExpansion`
        for available options
    weights_mlp
        configuration for the MLPs that map the radial basis functions
        to the weights of the interactions' tensor products
    channels
        the multiplicity of the node features corresponding to each irrep
        specified in ``hidden_irreps``
    hidden_irreps
        string representations of the :class:`e3nn.o3.Irrep`\ s to use
        for representing the node features between each message passing layer
    l_max
        the highest order to consider in:
        * the spherical harmonics expansion of the neighbour vectors
        * the irreps of node features used within each message passing layer
    layers
        number of message passing layers
    correlation
        maximum correlation (body-order) of the messages
    aggregation
        the type of aggregation to use when creating total messages from
        neigbour messages :math:`m_{j \rightarrow i}`
    self_connection
        whether to use self-connections in the message passing layers
    readout_width
        the width of the MLP used to read out the per-atom energies after the
        final message passing layer

    Examples
    --------
    Basic usage:

    .. code-block:: python

        >>> from graph_pes.models import MACE
        >>> model = MACE(
        ...     elements=["H", "C", "N", "O"],
        ...     cutoff=5.0,
        ...     channels=16,
        ...     radial_expansion="Bessel",
        ... )

    Specification in a YAML file:

    .. code-block:: yaml

        model:
            +MACE:
                elements: [H, C, N, O]
                cutoff: 5.0
                radial_expansion: Bessel

                # change from the default MLP config:
                weights_mlp:
                    hidden_depth: 2
                    hidden_features: 16
                    activation: SiLU

    """  # noqa: E501

    def __init__(
        self,
        elements: list[str],
        # radial things
        cutoff: float = DEFAULT_CUTOFF,
        n_radial: int = 8,
        radial_expansion: type[DistanceExpansion] | str = "Bessel",
        weights_mlp: MLPConfig = DEFAULT_MLP_CONFIG,
        # node things
        channels: int = 128,
        hidden_irreps: str | list[str] = "0e + 1o",
        # message passing things
        l_max: int = 3,
        layers: int = 2,
        correlation: int = 3,
        aggregation: NeighbourAggregationMode = "constant_fixed",
        self_connection: bool = True,
        # readout
        readout_width: int = 16,
    ):
        Z_embedding = AtomicOneHot(elements)
        Z_dim = len(elements)
        hidden_irrep_s = parse_irreps(hidden_irreps)
        nodes = NodeDescription(
            channels=channels,
            attributes=Z_dim,
            hidden_features=hidden_irrep_s,
        )

        super().__init__(
            cutoff=cutoff,
            n_radial=n_radial,
            radial_expansion=radial_expansion,
            weights_mlp={**DEFAULT_MLP_CONFIG, **weights_mlp},
            nodes=nodes,
            node_attribute_generator=Z_embedding,
            l_max=l_max,
            layers=layers,
            correlation=correlation,
            neighbour_aggregation=aggregation,
            use_self_connection=self_connection,
            readout_width=readout_width,
        )


class ZEmbeddingMACE(_BaseMACE):
    """
    A variant of MACE that uses a fixed-size (``z_embed_dim``) per-element
    embedding of the atomic numbers to condition the ``TensorProduct`` update
    in the residual connection of the message passing layers, as well as the
    contractions in the message passing layers.

    Please cite the following if you use this model in your research:

    .. code-block:: bibtex

        @misc{Batatia2022MACE,
            title = {
                MACE: Higher Order Equivariant Message Passing
                Neural Networks for Fast and Accurate Force Fields
            },
            author = {
                Batatia, Ilyes and Kov{\'a}cs, D{\'a}vid P{\'e}ter and
                Simm, Gregor N. C. and Ortner, Christoph and Cs{\'a}nyi, G{\'a}bor
            },
            year = {2022},
            doi = {10.48550/arXiv.2206.07697},
        }

    All paramters are identical to :class:`~graph_pes.models.MACE`, except for the following:

    - ``elements`` is not required or used here
    - ``z_embed_dim`` controls size of the per-element embedding
    """  # noqa: E501

    def __init__(
        self,
        z_embed_dim: int = 4,
        # radial things
        cutoff: float = DEFAULT_CUTOFF,
        n_radial: int = 8,
        radial_expansion: type[DistanceExpansion] | str = "Bessel",
        weights_mlp: MLPConfig = DEFAULT_MLP_CONFIG,
        # node things
        channels: int = 128,
        hidden_irreps: str | list[str] = "0e + 1o",
        # message passing things
        l_max: int = 3,
        layers: int = 2,
        correlation: int = 3,
        aggregation: NeighbourAggregationMode = "constant_fixed",
        self_connection: bool = True,
        # readout
        readout_width: int = 16,
    ):
        Z_embedding = PerElementEmbedding(z_embed_dim)
        hidden_irrep_s = parse_irreps(hidden_irreps)
        nodes = NodeDescription(
            channels=channels,
            attributes=z_embed_dim,
            hidden_features=hidden_irrep_s,
        )

        super().__init__(
            cutoff=cutoff,
            n_radial=n_radial,
            radial_expansion=radial_expansion,
            weights_mlp={**DEFAULT_MLP_CONFIG, **weights_mlp},
            nodes=nodes,
            node_attribute_generator=Z_embedding,
            l_max=l_max,
            layers=layers,
            correlation=correlation,
            neighbour_aggregation=aggregation,
            use_self_connection=self_connection,
            readout_width=readout_width,
        )



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Literal, Protocol, TypeVar

import dacite
import data2objects

from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models import AdditionModel
from graph_pes.training.loss import Loss, TotalLoss
from graph_pes.utils.misc import nested_merge

T = TypeVar("T")


def _nice_dict_repr(d: dict) -> str:
    def _print_dict(d: dict, indent: int = 0) -> str:
        nice = {
            k: v
            if not isinstance(v, dict)
            else "\n" + _print_dict(v, indent + 1)
            for k, v in d.items()
        }
        return "\n".join([f"{'  ' * indent}{k}: {v}" for k, v in nice.items()])

    return _print_dict(d, 0)


class HasDefaults(Protocol):
    @classmethod
    def defaults(cls) -> dict: ...


HD = TypeVar("HD", bound=HasDefaults)


def instantiate_config_from_dict(
    config_dict: dict, config_class: type[HD]
) -> tuple[dict, HD]:
    """Instantiate a config object from a dictionary."""

    config_dict = nested_merge(config_class.defaults(), config_dict)
    final_dict: dict = data2objects.fill_referenced_parts(config_dict)  # type: ignore

    import graph_pes
    import graph_pes.data
    import graph_pes.interfaces
    import graph_pes.models
    import graph_pes.training
    import graph_pes.training.callbacks
    import graph_pes.training.loss
    import graph_pes.training.opt

    object_dict = data2objects.from_dict(
        final_dict,
        modules=[
            graph_pes,
            graph_pes.models,
            graph_pes.training,
            graph_pes.training.opt,
            graph_pes.training.loss,
            graph_pes.data,
            graph_pes.training.callbacks,
            graph_pes.interfaces,
        ],
    )
    field_names = {f.name for f in fields(config_class)}  # type: ignore
    object_dict = {
        k: v
        for k, v in object_dict.items()
        if k in field_names  # type: ignore
    }

    try:
        return (
            final_dict,
            dacite.from_dict(
                data_class=config_class,
                data=object_dict,
                config=dacite.Config(strict=True),
            ),
        )
    except Exception as e:
        raise ValueError(
            f"Failed to instantiate a config object of type {config_class} "
            f"from the following dictionary:\n{_nice_dict_repr(final_dict)}"
        ) from e


@dataclass
class TorchConfig:
    """Configuration for PyTorch."""

    dtype: Literal["float16", "float32", "float64"]
    """
    The dtype to use for all model parameters and graph properties.
    Defaults is ``"float32"``.
    """

    float32_matmul_precision: Literal["highest", "high", "medium"]
    """
    The precision to use internally for float32 matrix multiplications. Refer to the
    `PyTorch documentation <https://pytorch.org/docs/stable/generated/torch.set_float32_matmul_precision.html>`__
    for details.

    Defaults to ``"high"`` to favour accelerated learning over numerical
    exactness for matmuls.
    """  # noqa: E501


def parse_model(
    model: GraphPESModel | dict[str, GraphPESModel],
) -> GraphPESModel:
    if isinstance(model, GraphPESModel):
        return model
    elif isinstance(model, dict):
        if not all(isinstance(m, GraphPESModel) for m in model.values()):
            _types = {k: type(v) for k, v in model.items()}

            raise ValueError(
                "Expected all values in the model dictionary to be "
                "GraphPESModel instances, but got something else: "
                f"types: {_types}\n"
                f"values: {model}\n"
            )
        return AdditionModel(**model)
    raise ValueError(
        "Expected to be able to parse a GraphPESModel or a "
        "dictionary of named GraphPESModels from the model config, "
        f"but got something else: {model}"
    )


def parse_loss(
    loss: Loss | TotalLoss | dict[str, Loss] | list[Loss],
) -> TotalLoss:
    if isinstance(loss, Loss):
        return TotalLoss([loss])
    elif isinstance(loss, TotalLoss):
        return loss
    elif isinstance(loss, dict):
        return TotalLoss(list(loss.values()))
    elif isinstance(loss, list):
        return TotalLoss(loss)
    raise ValueError(
        "Expected to be able to parse a Loss, TotalLoss, a list of "
        "Loss instances, or a dictionary mapping keys to Loss instances from "
        "the loss config, but got something else:\n"
        f"{loss}"
    )



from .datasets import (
    ASEToGraphDataset,
    ConcatDataset,
    DatasetCollection,
    GraphDataset,
    file_dataset,
    load_atoms_dataset,
)
from .loader import GraphDataLoader

__all__ = [
    "load_atoms_dataset",
    "file_dataset",
    "GraphDataset",
    "ASEToGraphDataset",
    "DatasetCollection",
    "GraphDataLoader",
    "ConcatDataset",
]



from __future__ import annotations

from typing import Iterator, Sequence, TypeVar, overload

import numpy as np

T = TypeVar("T")


class SequenceSampler(Sequence[T]):
    """
    A class that wraps a :class:`Sequence` of ``T`` objects and
    provides methods for sampling from it without the need to manipulate or
    access the underlying data.

    This is useful for e.g. sub-sampling a ``collection`` where individual
    indexing operations are expensive, such as a database on disk, or when
    indexing involves some form of pre-processing.

    Parameters
    ----------
    collection
        The collection to wrap.
    indices
        The indices of the elements to include in the collection. If ``None``,
        all elements are included.
    """

    def __init__(
        self,
        collection: Sequence[T],
        indices: Sequence[int] | None = None,
    ):
        self.collection = collection
        self.indices = indices or range(len(collection))

    @overload
    def __getitem__(self, index: int) -> T: ...
    @overload
    def __getitem__(self, index: slice) -> SequenceSampler[T]: ...
    def __getitem__(self, index: int | slice) -> T | SequenceSampler[T]:
        """
        Get the element/s at the given ``index``.

        Parameters
        ----------
        index
            The index/indices of the element/elements to get.
        """
        if isinstance(index, int):
            return self.collection[self.indices[index]]

        sampled_indices = self.indices[index]
        return SequenceSampler(self.collection, sampled_indices)

    def __len__(self) -> int:
        """The number of items in the collection."""
        return len(self.indices)

    def __iter__(self) -> Iterator[T]:
        """Iterate over the collection."""
        for i in range(len(self)):
            yield self[i]

    def shuffled(self, seed: int = 42) -> SequenceSampler[T]:
        """
        Return a shuffled version of this collection.

        Parameters
        ----------
        seed
            The random seed to use for shuffling.
        """
        # 1. make a copy of the indices
        indices = [*self.indices]
        # 2. shuffle them
        np.random.default_rng(seed).shuffle(indices)
        # 3. return a new OrderedCollection with the shuffled indices
        return SequenceSampler(self.collection, indices)

    def sample_at_most(self, n: int, seed: int = 42) -> SequenceSampler[T]:
        """
        Return a sampled collection with at most ``n`` elements.
        """
        assert n >= 0, "n must be non-negative"
        n = min(n, len(self))
        return self.shuffled(seed)[:n]



from __future__ import annotations

import torch

from graph_pes.atomic_graph import (
    DEFAULT_CUTOFF,
    AtomicGraph,
    PropertyKey,
    index_over_neighbours,
    neighbour_distances,
    sum_over_neighbours,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.components.scaling import LocalEnergiesScaler
from graph_pes.utils.misc import uniform_repr
from graph_pes.utils.nn import (
    MLP,
    PerElementEmbedding,
    ShiftedSoftplus,
    UniformModuleList,
)

from .components.distances import DistanceExpansion, GaussianSmearing


class CFConv(torch.nn.Module):
    r"""
    CFConv: The Continous Filter Convolution

    Originally proposed as part of `SchNet <https://arxiv.org/abs/1706.08566>`_,
    this message passing layer creates messages of the form:

    .. math::
        m_{ij} = x_j \odot \mathbb{F}(r_{ij})

    where :math:`\mathbb{F}` is some learnable filter function which
    expands the radial distance :math:`r_{ij}` into a set of features
    that match the dimensionality of the node features :math:`x_j`.

    Example
    -------
    .. code::

        from graph_pes.models.components.distances import GaussianSmearing
        from graph_pes.utils.nn import MLP
        from torch import nn

        filter_generator = nn.Sequential(
            GaussianSmearing(8, 5.0),
            MLP([8, 16, 16, 8]),
        )
        cfconv = CFConv(filter_generator)

    Parameters
    ----------
    filter_generator
        The filter function :math:`\mathbb{F}`.
    """

    def __init__(self, filter_generator: torch.nn.Module):
        super().__init__()
        self.filter_generator = filter_generator

    def forward(
        self,
        channels: torch.Tensor,  # (n_atoms, F)
        edge_distances: torch.Tensor,  # (E,)
        graph: AtomicGraph,
    ) -> torch.Tensor:  # (n_atoms, F)
        edge_features = self.filter_generator(edge_distances)  # (E, F)
        neighbour_features = index_over_neighbours(channels, graph)  # (E, F)

        messages = neighbour_features * edge_features  # (E, F)

        return sum_over_neighbours(messages, graph)

    def __repr__(self):
        rep = f"CFConv({self.filter_generator})"
        if len(rep) > 40:
            indented = "  " + "\n  ".join(
                str(self.filter_generator).split("\n")
            )
            rep = f"CFConv(\n{indented}\n)"
        return rep


class SchNetInteraction(torch.nn.Module):
    r"""
    SchNet interaction block.

    Updates the embedding of each atom, :math:`x_i` by sequentially
    applying the following:

    1. a linear transform of each node's features :math:`x_i \leftarrow W x_i`
    2. message creation :math:`m_{ij} = x_j \odot \mathbb{F}(r_{ij})`
    3. message aggregation
       :math:`m_i = \sum_{j \in \mathcal{N}(i)} m_{ij}`
    4. a multi-layer perceptron that further embeds these new node features
       :math:`x_i \leftarrow \mathrm{MLP}(h_i)`

    Parameters
    ----------
    channels
        Number of channels of the internal representations.
    expansion_features
        Number of features used for the radial basis expansion.
    cutoff
        Neighborhood cutoff radius.
    basis_type
        The type of radial basis expansion to use.
    """

    def __init__(
        self,
        channels: int,
        expansion_features: int,
        cutoff: float,
        basis_type: type[DistanceExpansion],
    ):
        super().__init__()

        # schnet interaction block's are composed of 3 elements

        # 1. linear transform to get new node features
        self.linear = torch.nn.Linear(channels, channels, bias=False)

        # 2. cfconv to mix these new features with distances information,
        # and aggregate over neighbors to create completely new node features
        filter_generator = torch.nn.Sequential(
            basis_type(expansion_features, cutoff),
            MLP(
                [expansion_features, channels, channels],
                activation=ShiftedSoftplus(),
            ),
        )
        self.cfconv = CFConv(filter_generator)

        # 3. mlp to further embed these new node features
        self.mlp = MLP(
            [channels, channels, channels],
            activation=ShiftedSoftplus(),
        )

    def forward(
        self,
        features: torch.Tensor,
        neighbour_distances: torch.Tensor,
        graph: AtomicGraph,
    ):
        # 1. linear transform to get new node features
        h = self.linear(features)

        # 2. cfconv to mix these new features with distances information,
        # and aggregate over neighbors to create completely new node features
        h = self.cfconv(h, neighbour_distances, graph)

        # 3. mlp to further embed these new node features
        return self.mlp(h)


class SchNet(GraphPESModel):
    r"""
    The `SchNet <https://arxiv.org/abs/1706.08566>`_ model: a pairwise, scalar,
    message passing GNN.

    A stack of :class:`~graph_pes.models.schnet.SchNetInteraction` blocks are
    used to update the node features of each atom, :math:`x_i` by sequentially
    aggregating over neighbouring atom's features, :math:`x_j`:, and
    distances, :math:`r_{ij}`:.

    Citation:

    .. code:: bibtex

        @article{Schutt-18-03,
            title = {{{SchNet}} {\textendash} {{A}} Deep Learning Architecture for Molecules and Materials},
            author = {Sch{\"u}tt, K. T. and Sauceda, H. E. and Kindermans, P.-J. and Tkatchenko, A. and M{\"u}ller, K.-R.},
            year = {2018},
            journal = {The Journal of Chemical Physics},
            volume = {148},
            number = {24},
            pages = {241722},
            doi = {10.1063/1.5019779},
        }

    Parameters
    ----------
    channels
        Number of features per node.
    expansion_features
        Number of features used for the radial basis expansion.
    cutoff
        Neighborhood cutoff radius.
    layers
        Number of interaction blocks to apply.
    expansion
        The type of radial basis expansion to use. Defaults to
        :class:`GaussianSmearing <graph_pes.models.components.distances.GaussianSmearing>`
        as in the original paper.
    """  # noqa: E501

    def __init__(
        self,
        cutoff: float = DEFAULT_CUTOFF,
        channels: int = 64,
        expansion_features: int = 50,
        layers: int = 3,
        expansion: type[DistanceExpansion] | None = None,
    ):
        super().__init__(
            cutoff=cutoff,
            implemented_properties=["local_energies"],
        )

        if expansion is None:
            expansion = GaussianSmearing

        self.chemical_embedding = PerElementEmbedding(channels)

        self.interactions = UniformModuleList(
            SchNetInteraction(channels, expansion_features, cutoff, expansion)
            for _ in range(layers)
        )

        self.read_out = MLP(
            [channels, channels // 2, 1],
            activation=ShiftedSoftplus(),
        )

        self.scaler = LocalEnergiesScaler()

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        h = self.chemical_embedding(graph.Z)
        d = neighbour_distances(graph)

        for interaction in self.interactions:
            h = h + interaction(h, d, graph)

        local_energies = self.read_out(h).squeeze()
        local_energies = self.scaler(local_energies, graph)

        return {"local_energies": local_energies}

    def __repr__(self) -> str:
        return uniform_repr(
            self.__class__.__name__,
            chemical_embedding=self.chemical_embedding,
            interactions=self.interactions,
            read_out=self.read_out,
            scaler=self.scaler,
        )



from __future__ import annotations

import ase.build
import numpy as np
import pytest
import torch
from e3nn import o3
from mace.calculators import MACECalculator
from mace.modules import ScaleShiftMACE, gate_dict, interaction_classes

from graph_pes.atomic_graph import AtomicGraph, to_batch
from graph_pes.interfaces._mace import (
    MACEWrapper,
    go_mace_23,
    mace_mp,
    mace_off,
)
from graph_pes.utils.calculator import GraphPESCalculator

ELEMENTS = [1, 6, 8]
CUTOFF = 5.0


def create_default_scaleshift_mace(
    atomic_numbers: list,
    r_max: float = 5.0,
    num_bessel: int = 8,
    num_polynomial_cutoff: int = 5,
    max_ell: int = 3,
    num_interactions: int = 3,
    hidden_irreps: str = "32x0e + 32x1o",
    mlp_irreps: str = "16x0e",
    atomic_energies: torch.Tensor | None = None,
    avg_num_neighbors: float = 1.0,
    atomic_inter_scale: float = 1.0,
    atomic_inter_shift: float = 0.0,
) -> ScaleShiftMACE:
    if atomic_energies is None:
        atomic_energies = torch.zeros(len(atomic_numbers))

    interaction_cls = interaction_classes[
        "RealAgnosticResidualInteractionBlock"
    ]
    interaction_cls_first = interaction_classes["RealAgnosticInteractionBlock"]

    # Default gate function
    gate = gate_dict["silu"]

    # Create model
    model = ScaleShiftMACE(
        r_max=r_max,
        num_bessel=num_bessel,
        num_polynomial_cutoff=num_polynomial_cutoff,
        max_ell=max_ell,
        num_interactions=num_interactions,
        num_elements=len(atomic_numbers),
        hidden_irreps=o3.Irreps(hidden_irreps),
        MLP_irreps=o3.Irreps(mlp_irreps),
        atomic_energies=atomic_energies,
        avg_num_neighbors=avg_num_neighbors,
        atomic_numbers=atomic_numbers,
        correlation=3,
        atomic_inter_scale=atomic_inter_scale,
        atomic_inter_shift=atomic_inter_shift,
        # Added required parameters
        interaction_cls=interaction_cls,
        interaction_cls_first=interaction_cls_first,
        gate=gate,
        # Optional parameters with defaults
        pair_repulsion=False,
        distance_transform="None",
        radial_MLP=[64, 64, 64],
        radial_type="bessel",
        heads=None,
        cueq_config=None,
    )

    return model


MACE_MODEL = create_default_scaleshift_mace(ELEMENTS, CUTOFF)

# Test molecules/crystals
CH4 = ase.build.molecule("CH4")
DIAMOND = ase.build.bulk("C", "diamond", a=3.5668)

# Pre-configured calculators
MACE_CALC = MACECalculator(models=MACE_MODEL)
GRAPH_PES_MODEL = MACEWrapper(MACE_MODEL)
GRAPH_PES_CALC = GraphPESCalculator(GRAPH_PES_MODEL)


def test_output_shapes():
    graph = AtomicGraph.from_ase(DIAMOND)
    outputs = GRAPH_PES_MODEL.get_all_PES_predictions(graph)

    assert outputs["energy"].shape == ()
    assert outputs["forces"].shape == (2, 3)  # 2 atoms in unit cell
    assert outputs["stress"].shape == (3, 3)

    batch = to_batch([graph, graph])
    outputs = GRAPH_PES_MODEL.get_all_PES_predictions(batch)

    assert outputs["energy"].shape == (2,)
    assert outputs["forces"].shape == (4, 3)
    assert outputs["stress"].shape == (2, 3, 3)


def test_molecular():
    MACE_CALC.calculate(CH4, properties=["energy", "forces", "stress"])
    GRAPH_PES_CALC.calculate(CH4, properties=["energy", "forces", "stress"])

    assert MACE_CALC.results["energy"] == pytest.approx(
        GRAPH_PES_CALC.results["energy"]
    )
    np.testing.assert_allclose(
        MACE_CALC.results["forces"],
        GRAPH_PES_CALC.results["forces"],
        atol=1e-4,
        rtol=100,
    )
    np.testing.assert_allclose(
        MACE_CALC.results["stress"],
        GRAPH_PES_CALC.results["stress"],
        atol=1e-4,
        rtol=100,
    )


def test_periodic():
    MACE_CALC.calculate(DIAMOND, properties=["energy", "forces", "stress"])
    GRAPH_PES_CALC.calculate(DIAMOND, properties=["energy", "forces", "stress"])

    assert MACE_CALC.results["energy"] == pytest.approx(
        GRAPH_PES_CALC.results["energy"], abs=1e-4
    )
    np.testing.assert_allclose(
        MACE_CALC.results["forces"],
        GRAPH_PES_CALC.results["forces"],
        atol=1e-4,
        rtol=100,
    )
    np.testing.assert_allclose(
        MACE_CALC.results["stress"].flatten(),
        GRAPH_PES_CALC.results["stress"].flatten(),
        atol=1e-4,
        rtol=100,
    )


def test_mace_mp():
    # check can download model
    base_model = mace_mp("small", "float32")
    calc = GraphPESCalculator(base_model)

    # check that forces on central atom are roughly zero
    calc.calculate(CH4, properties=["energy", "forces"])
    assert np.abs(calc.results["forces"][0]).max() < 1e-4


def test_mace_off():
    base_model = mace_off("small", "float32")
    calc = GraphPESCalculator(base_model)

    # check that forces on central atom are roughly zero
    calc.calculate(CH4, properties=["energy", "forces"])
    assert np.abs(calc.results["forces"][0]).max() < 1e-4


def test_go_mace_23():
    base_model = go_mace_23()
    calc = GraphPESCalculator(base_model)

    calc.calculate(CH4, properties=["energy", "forces"])
    assert np.abs(calc.results["forces"][0]).max() < 1e-4



from __future__ import annotations

import torch

from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.offsets import LearnableOffset
from graph_pes.utils.misc import contains_tensor, uniform_repr


class Optimizer:
    """
    A factory class for delayed instantiation of :class:`torch.optim.Optimizer`
    objects.

    The generated optimizer splits the parameters of the model into two groups:

    - "non-decayable" parameters, which are all parameters returned by the
      :meth:`~graph_pes.GraphPESModel.non_decayable_parameters` method of
      the model.
    - "normal" parameters, corresponding to the remaining model parameters.

    Unsurprisingly, any specified weight decay is applied only to the normal
    model parameters.

    As an example, per-element energy offsets parameters of
    :class:`~graph_pes.models.offsets.LearnableOffset` models
    represent the arbitrary zero points of energies for different elements:
    it doesn't make sense to push these towards zero during training.

    .. note::

        We use delayed instantiation of optimizers when configuring our training
        runs to allow for arbitrary changes to the model and its parameters
        during the :class:`~graph_pes.GraphPESModel.pre_fit_all_components`
        method.


    Parameters
    ----------
    name
        The name of the :class:`torch.optim.Optimizer` class to use, e.g.
        ``"Adam"`` or ``"SGD"``. Alternatively, provide the type of any
        subclass of :class:`torch.optim.Optimizer`.
    **kwargs
        Additional keyword arguments to pass to the specified optimizer's
        constructor.

    Examples
    --------
    Pass a named optimiser:

    >>> from graph_pes.training.opt import Optimizer
    >>> optimizer_factory = Optimizer("AdamW", lr=1e-3)
    >>> optimizer_instance = optimizer_factory(model)

    Or pass the optimiser class directly:

    >>> from torch.optim import SGD
    >>> optimizer_factory = Optimizer(SGD, lr=1e-3)
    >>> optimizer_instance = optimizer_factory(model)

    Psuedo-code excerpt from ``graph-pes-train`` logic:

    >>> from graph_pes.training.opt import Optimizer
    >>> from graph_pes.models import LennardJones
    >>> ...
    >>> optimizer_factory = Optimizer("AdamW", lr=1e-3)
    >>> model = LennardJones()
    >>> model.pre_fit(train_loader)
    >>> optimizer_instance = optimizer_factory(model)
    """

    def __init__(
        self,
        name: str | type[torch.optim.Optimizer],
        **kwargs,
    ):
        self.kwargs = kwargs
        if isinstance(name, str):
            optimizer_class: type[torch.optim.Optimizer] = getattr(
                torch.optim, name, None
            )  # type: ignore
            if optimizer_class is not None:
                name = optimizer_class
            else:
                raise ValueError(
                    f"Could not find optimizer {name}. "
                    "Please provide the common PyTorch name, or a fully "
                    "qualified class name."
                )
        else:
            optimizer_class = name

        dummy_model = LearnableOffset()
        dummy_opt = optimizer_class(dummy_model.parameters(), **kwargs)
        if not isinstance(dummy_opt, torch.optim.Optimizer):
            raise ValueError(
                "Expected the returned optimizer to be an instance of "
                "torch.optim.Optimizer, but got "
                f"{type(dummy_opt)}"
            )

        self.optimizer_class = optimizer_class

    def __call__(self, model: GraphPESModel) -> torch.optim.Optimizer:
        """
        Create an instance of the specified optimizer class, with the correct
        parameter groups for the model.
        """

        all_params = list(set(model.parameters()))
        non_decayable_params = list(set(model.non_decayable_parameters()))

        param_groups = [
            {
                "name": "non-decayable",
                "params": non_decayable_params,
                "weight_decay": 0.0,
            },
            {
                "name": "normal",
                "params": [
                    p
                    for p in all_params
                    if not contains_tensor(non_decayable_params, p)
                ],
            },
        ]

        # remove empty groups
        param_groups = [pg for pg in param_groups if pg["params"]]

        return self.optimizer_class(param_groups, **self.kwargs)

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            name=self.optimizer_class.__name__,
            **self.kwargs,
        )


class LRScheduler:
    """
    A factory class for delayed instantiation of
    :class:`torch.optim.lr_scheduler.LRScheduler` objects.

    Parameters
    ----------
    name
        The name of the :class:`torch.optim.lr_scheduler.LRScheduler` class to
        use, e.g. ``"ReduceLROnPlateau"``. Alternatively, provide any subclass
        of :class:`torch.optim.lr_scheduler.LRScheduler`.
    **kwargs
        Additional keyword arguments to pass to the specified scheduler's
        constructor.

    Examples
    --------
    >>> from graph_pes.training.opt import LRScheduler
    >>> ...
    >>> scheduler_factory = LRScheduler(
    ...     "LambdaLR", lr_lambda=lambda epoch: 0.95 ** epoch
    ... )
    >>> scheduler_instance = scheduler_factory(optimizer)
    """

    def __init__(
        self,
        name: str | type[torch.optim.lr_scheduler.LRScheduler],
        **kwargs,
    ):
        self.kwargs = kwargs
        if isinstance(name, str):
            scheduler_class: type[torch.optim.lr_scheduler.LRScheduler] = (
                getattr(torch.optim.lr_scheduler, name, None)
            )  # type: ignore
            if scheduler_class is not None:
                name = scheduler_class
            else:
                raise ValueError(
                    f"Could not find scheduler {name}. "
                    "Please provide the common PyTorch name, or a fully "
                    "qualified class name."
                )
        else:
            scheduler_class = name

        dummy_opt = torch.optim.Adam(LearnableOffset().parameters())
        dummy_scheduler = scheduler_class(dummy_opt, **kwargs)
        if not isinstance(
            dummy_scheduler,
            (
                torch.optim.lr_scheduler.LRScheduler,
                torch.optim.lr_scheduler.ReduceLROnPlateau,
            ),
        ):
            raise ValueError(
                "Expected the returned scheduler to be an instance of "
                "torch.optim.lr_scheduler.LRScheduler, but got "
                f"{type(dummy_scheduler)}"
            )

        self.scheduler_class = scheduler_class

    def __call__(
        self, optimizer: torch.optim.Optimizer
    ) -> torch.optim.lr_scheduler.LRScheduler:
        return self.scheduler_class(optimizer, **self.kwargs)

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            name=self.scheduler_class.__name__,
            **self.kwargs,
        )



from __future__ import annotations

from typing import Callable, Iterable

import ase
import matplotlib.axes
import matplotlib.lines
import matplotlib.pyplot as plt
import numpy as np
import torch
from cycler import cycler
from matplotlib.ticker import MaxNLocator
from torch import Tensor

from graph_pes.utils.calculator import GraphPESCalculator, merge_predictions
from graph_pes.utils.misc import voigt_6_to_full_3x3

from ..atomic_graph import AtomicGraph, PropertyKey, to_batch
from ..graph_pes_model import GraphPESModel

Transform = Callable[[Tensor, AtomicGraph], Tensor]
r"""
Transforms map a property, :math:`x`, to a target property, :math:`y`,
conditioned on an :class:`~graph_pes.AtomicGraph`, :math:`\mathcal{G}`:

.. math::

    T: (x; \mathcal{G}) \mapsto y
"""


def identity(x: Tensor, graph: AtomicGraph) -> Tensor:
    return x


_my_style = {
    "figure.figsize": (3.5, 3),
    "axes.spines.right": False,
    "axes.spines.top": False,
    "legend.frameon": False,
    "legend.fancybox": False,
    "savefig.transparent": False,
    "axes.prop_cycle": cycler(
        "color",
        [
            "#4385be",
            "#ce5d97",
            "#da702c",
            "#8b7ec8",
            "#879a39",
            "#d14d41",
            "#d14d41",
            "#d0a215",
            "#3aa99f",
        ],
    ),
}

plt.rcParams.update(_my_style)


def move_axes(ax: matplotlib.axes.Axes | None = None):  # type: ignore
    """Move the axes outward."""
    ax: plt.Axes = ax or plt.gca()
    ax.spines["left"].set_position(("outward", 10))
    ax.spines["bottom"].set_position(("outward", 10))


def parity_plot(
    model: GraphPESModel | GraphPESCalculator,
    structures: Iterable[AtomicGraph] | Iterable[ase.Atoms],
    property: PropertyKey = "energy",
    transform: Transform | None = None,
    units: str | None = None,
    ax: matplotlib.axes.Axes | None = None,  # type: ignore
    batch_size: int = 5,
    **scatter_kwargs,
):
    r"""
    A nicely formatted parity plot of model predictions vs ground truth
    for the given :code:`property`.

    Parameters
    ----------
    model
        The model to for generating predictions.
    structures
        The structures to make predictions on.
    property
        The property to plot, e.g. :code:`"energy"`.
    transform
        The transform to apply to the predictions and labels before plotting.
        If not provided, no transform is applied.
    units
        The units of the property, for labelling the axes. If not provided, no
        units are used.
    ax
        The axes to plot on. If not provided, the current axes are used.
    batch_size
        The size of the batch to use for making predictions.
    scatter_kwargs
        Keyword arguments to pass to :code:`plt.scatter`.

    Examples
    --------
    Default settings (no units, transforms or custom scatter keywords):

    .. code-block:: python

        parity_plot(model, train, "energy")

    .. image:: Cu-LJ-default-parity.svg
        :align: center

    Custom settings, as seen in
    :doc:`this example notebook <../quickstart/custom-training-loop>`:

    .. code-block:: python

        from graph_pes.atomic_graph import DividePerAtom
        from graph_pes.util import Keys

        for name, data, colour in zip(
            ["Train", "Test"],
            [train, test],
            ["royalblue", "crimson"],
        ):
            parity_plot(
                model,
                data,
                "energy",
                transform=DividePerAtom(),
                units="eV / atom",
                label=name,
                c=colour,
            )

        plt.legend(loc="upper left", fancybox=False);

    .. image:: ../quickstart/parity-plot.svg
        :align: center
    """
    # deal with defaults
    transform = transform or identity
    calc = (
        GraphPESCalculator(model) if isinstance(model, GraphPESModel) else model
    )

    graphs = [
        AtomicGraph.from_ase(s, calc.model.cutoff.item() + 0.001)
        if isinstance(s, ase.Atoms)
        else s
        for s in structures
    ]

    # get the predictions
    per_struct_predictions = calc.calculate_all(graphs, [property], batch_size)
    if any(property not in g.properties for g in graphs):
        raise ValueError(
            f"Property {property} is not available for all structures "
            "you passed"
        )

    predictions = torch.tensor(
        merge_predictions(per_struct_predictions)[property]
    )
    if property in ["stress", "virial"]:
        # reconvert from calculator ase format to 3x3
        predictions = voigt_6_to_full_3x3(predictions)

    # okay to form a big batch since not passing through model
    batch = to_batch(graphs)

    # transform
    ground_truth = transform(batch.properties[property], batch).detach()
    predicted = transform(predictions, batch)

    # plot
    ax: plt.Axes = ax or plt.gca()

    default_kwargs = dict(lw=0, clip_on=False)
    scatter_kwargs = {**default_kwargs, **scatter_kwargs}
    ax.scatter(ground_truth, predicted, **scatter_kwargs)

    # get a point guaranteed to be on the plot
    z = ground_truth.view(-1)[0].item()
    ax.axline((z, z), slope=1, c="k", ls="--", lw=1)

    # aesthetics
    axis_label = (
        f"{property.capitalize()} ({units})" if units else property.capitalize()
    )
    ax.set_xlabel(f"True {axis_label}")
    ax.set_ylabel(f"Predicted {axis_label}")
    ax.set_aspect("equal", "datalim")
    x0, x1 = ax.get_xlim()
    y0, y1 = ax.get_ylim()
    ax.set_xlim(min(x0, y0), max(x1, y1))
    ax.set_ylim(min(x0, y0), max(x1, y1))
    move_axes(ax)

    # 5 ticks each
    ax.xaxis.set_major_locator(MaxNLocator(5))
    ax.yaxis.set_major_locator(MaxNLocator(5))


def dimer_curve(
    model: GraphPESModel,
    system: str,
    units: str | None = None,
    set_to_zero: bool = True,
    rmin: float = 0.9,
    rmax: float | None = None,
    ax: matplotlib.axes.Axes | None = None,  # type: ignore
    auto_lim: bool = True,
    **plot_kwargs,
) -> matplotlib.lines.Line2D:
    r"""
    A nicely formatted dimer curve plot for the given :code:`system`.

    Parameters
    ----------
    model
        The model for generating predictions.
    system
        The dimer system. Should be one of: a single element, e.g. :code:`"Cu"`,
        or a pair of elements, e.g. :code:`"CuO"`.
    units
        The units of the energy, for labelling the axes. If not provided, no
        units are used.
    set_to_zero
        Whether to set the energy of the dimer at :code:`rmax` to be zero.
    rmin
        The minimum seperation to consider.
    rmax
        The maximum seperation to consider.
    ax
        The axes to plot on. If not provided, the current axes are used.
    plot_kwargs
        Keyword arguments to pass to :code:`plt.plot`.

    Examples
    --------

    .. code-block:: python

        from graph_pes.utils.analysis import dimer_curve
        from graph_pes.models import LennardJones

        dimer_curve(LennardJones(sigma=1.3, epsilon=0.5), system="OH", units="eV")

    .. image:: dimer-curve.svg
        :align: center
    """  # noqa: E501

    trial_atoms = ase.Atoms(system)
    if len(trial_atoms) != 2:
        system = system + "2"

    if rmax is None:
        rmax = model.cutoff.item() + 0.5
    rs = np.linspace(rmin, rmax, 200)
    dimers = [ase.Atoms(system, positions=[[0, 0, 0], [r, 0, 0]]) for r in rs]
    graphs = [AtomicGraph.from_ase(d, cutoff=rmax + 0.1) for d in dimers]
    batch = to_batch(graphs)

    with torch.no_grad():
        energy = model.predict_energy(batch).numpy()

    if set_to_zero:
        energy -= energy[-1]

    ax: plt.Axes = ax or plt.gca()

    default_kwargs = dict(lw=1, c="k")
    plot_kwargs = {**default_kwargs, **plot_kwargs}
    line = ax.plot(rs, energy, **plot_kwargs)[0]
    assert isinstance(line, matplotlib.lines.Line2D)

    if auto_lim:
        limiting_energy = energy[-1]
        if (energy[:-1] < limiting_energy).any():
            well_depth = limiting_energy - energy[:-1].min()
        else:
            well_depth = 0.1
        bottom = limiting_energy - well_depth * 1.1
        top = limiting_energy + well_depth * 1.1
        ax.set_ylim(bottom, top)

        first_in_view = np.where(energy < top)[0][0]
        ax.set_xlim(rs[first_in_view].item() - 0.2, rs[-1] + 0.2)

    ax.set_xlabel("r (Å)")
    ax.set_ylabel(f"Dimer Energy ({units})" if units else "Dimer Energy")

    move_axes(ax)

    # 5 ticks each
    ax.xaxis.set_major_locator(MaxNLocator(5))
    ax.yaxis.set_major_locator(MaxNLocator(5))

    return line



import argparse
import os
from datetime import datetime
from pathlib import Path

import yaml

from graph_pes.config.shared import (
    instantiate_config_from_dict,
    parse_loss,
    parse_model,
)
from graph_pes.config.training import TrainingConfig
from graph_pes.data.loader import GraphDataLoader
from graph_pes.scripts.train import trainer_from_config
from graph_pes.training.tasks import TrainingTask
from graph_pes.utils.logger import logger


def parse_args():
    parser = argparse.ArgumentParser(
        description="Resume a `graph-pes-train` training run.",
        epilog="Copyright 2024, John Gardner",
    )
    parser.add_argument(
        "train_directory",
        type=str,
        help=(
            "Path to the training directory. For instance, "
            "`graph-pes-results/abdcefg_hijklmn`"
        ),
    )

    return parser.parse_args()


def main():
    # set the load-atoms verbosity to 1 by default to avoid
    # spamming logs with `rich` output
    os.environ["LOAD_ATOMS_VERBOSE"] = os.getenv("LOAD_ATOMS_VERBOSE", "1")

    args = parse_args()

    train_dir = Path(args.train_directory)
    if not train_dir.exists():
        raise ValueError(f"Training directory not found: {train_dir}")

    # find the latest checkpoint
    checkpoint_path = train_dir / "checkpoints/last.ckpt"
    assert checkpoint_path.exists(), f"Checkpoint not found: {checkpoint_path}"

    # and the training config
    config_path = train_dir / "train-config.yaml"
    assert config_path.exists(), f"Training config not found: {config_path}"

    with open(config_path) as f:
        config_data = yaml.safe_load(f)

    # load the checkpoint
    config_data, config = instantiate_config_from_dict(
        config_data, TrainingConfig
    )
    task = TrainingTask.load_from_checkpoint(
        checkpoint_path,
        model=parse_model(config.model),
        loss=parse_loss(config.loss),
        optimizer=config.fitting.optimizer,
        scheduler=config.fitting.scheduler,
    )

    # create the trainer
    trainer = trainer_from_config(config, train_dir)
    if trainer.global_rank == 0:
        now_ms = datetime.now().strftime("%F %T.%f")[:-3]
        logger.info(f"Resuming training at {now_ms}")

    # resume training
    data = config.get_data()
    loader_kwargs = {**config.fitting.loader_kwargs}
    loader_kwargs["shuffle"] = True
    train_loader = GraphDataLoader(data.train, **loader_kwargs)
    loader_kwargs["shuffle"] = False
    valid_loader = GraphDataLoader(data.valid, **loader_kwargs)
    trainer.fit(task, train_loader, valid_loader, ckpt_path=checkpoint_path)


if __name__ == "__main__":
    main()



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Iterable, Literal

import pytorch_lightning as pl
import torch
from pytorch_lightning.utilities.types import OptimizerLRSchedulerConfig

from graph_pes.atomic_graph import (
    AtomicGraph,
    PropertyKey,
    number_of_structures,
    to_batch,
)
from graph_pes.config.training import FittingOptions
from graph_pes.data.datasets import DatasetCollection, GraphDataset
from graph_pes.data.loader import GraphDataLoader
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.training.loss import (
    MAE,
    RMSE,
    Loss,
    PerAtomEnergyLoss,
    PropertyLoss,
    TotalLoss,
)
from graph_pes.training.opt import LRScheduler, Optimizer
from graph_pes.training.utils import (
    VALIDATION_LOSS_KEY,
    log_model_info,
    sanity_check,
)
from graph_pes.utils.logger import logger
from graph_pes.utils.nn import PerElementParameter
from graph_pes.utils.sampling import SequenceSampler


def train_with_lightning(
    trainer: pl.Trainer,
    model: GraphPESModel,
    data: DatasetCollection,
    loss: TotalLoss,
    fit_config: FittingOptions,
    optimizer: Optimizer,
    user_eval_metrics: list[Loss] | None = None,
    scheduler: LRScheduler | None = None,
):
    # - prepare the data
    if trainer.global_rank == 0:
        logger.info("Preparing data")
        data.train.prepare_data()
        data.valid.prepare_data()
    trainer.strategy.barrier("data prepare")

    logger.info("Setting up datasets")
    data.train.setup()
    data.valid.setup()

    loader_kwargs = {**fit_config.loader_kwargs}
    loader_kwargs["shuffle"] = True
    train_loader = GraphDataLoader(data.train, **loader_kwargs)
    loader_kwargs["shuffle"] = False
    valid_loader = GraphDataLoader(data.valid, **loader_kwargs)

    eval_metrics = get_all_eval_metrics(
        [data.train, data.valid], user_eval_metrics
    )

    # - do some pre-fitting
    pre_fit_graphs = SequenceSampler(data.train.graphs)
    if fit_config.max_n_pre_fit is not None:
        pre_fit_graphs = pre_fit_graphs.sample_at_most(fit_config.max_n_pre_fit)
    pre_fit_graphs = list(pre_fit_graphs)

    # optionally pre-fit the model
    if fit_config.pre_fit_model:
        logger.info(f"Pre-fitting the model on {len(pre_fit_graphs):,} samples")
        model.pre_fit_all_components(pre_fit_graphs)
    trainer.strategy.barrier("pre-fit")

    # always register the elements in the training set
    for param in model.parameters():
        if isinstance(param, PerElementParameter):
            param.register_elements(
                torch.unique(to_batch(pre_fit_graphs).Z).tolist()
            )

    # always pre-fit the losses
    for subloss in loss.losses:
        subloss.pre_fit(to_batch(pre_fit_graphs))

    # - log the model info
    if trainer.global_rank == 0:
        log_model_info(model, trainer.logger)

    # - sanity checks
    logger.info("Sanity checking the model...")
    sanity_check(model, next(iter(train_loader)))

    # - create the task (a pytorch lightning module)
    task = TrainingTask(model, loss, optimizer, scheduler, eval_metrics)

    # - train the model
    logger.info("Starting fit...")
    try:
        trainer.fit(task, train_loader, valid_loader)
    except Exception as e:
        logger.error(f"Training failed: {e}")
        raise e
    finally:
        try:
            task.load_best_weights(model, trainer)
        except Exception as e:
            logger.error(f"Failed to load best weights: {e}")


class TrainingTask(pl.LightningModule):
    def __init__(
        self,
        model: GraphPESModel,
        loss: TotalLoss,
        optimizer: Optimizer,
        scheduler: LRScheduler | None,
        eval_metrics: list[Loss],
    ):
        super().__init__()
        self.model = model
        self.optimizer_factory = optimizer
        self.scheduler_factory = scheduler
        self.total_loss = loss
        self.train_properties: list[PropertyKey] = list(
            set.union(
                set(),
                *[
                    set(loss.required_properties)
                    for loss in self.total_loss.losses
                ],
            )
        )
        self.eval_metrics = eval_metrics
        self.validation_properties = list(
            set.union(
                set(), *[set(m.required_properties) for m in eval_metrics]
            )
        )

    def forward(self, graphs: AtomicGraph) -> torch.Tensor:
        """Get the energy"""
        return self.model.predict_energy(graphs)

    def on_train_start(self):
        super().on_train_start()
        self.model.train()

    def on_validation_start(self):
        super().on_validation_start()
        self.model.eval()

    def on_validation_end(self) -> None:
        super().on_validation_end()
        self.model.train()

    def _step(
        self,
        graph: AtomicGraph,
        mode: Literal["train", "valid"],
    ) -> torch.Tensor:
        """
        Get (and log) the losses and metrics for a step.

        Parameters
        ----------
        graph: AtomicGraph
            The atomic graph to compute the losses and metrics for.
        prefix: Literal["train", "valid", "test"]
            The prefix of the step, either "train", "valid", or "test".

        Returns
        -------
        torch.Tensor
            The total loss for the step.
        """

        def log(name: str, value: torch.Tensor | float):
            if isinstance(value, torch.Tensor):
                value = value.item()

            validating = mode == "valid"

            return self.log(
                f"{mode}/{name}",
                value,
                prog_bar=validating and "metric" in name,
                on_step=not validating,
                on_epoch=validating,
                sync_dist=validating,
                batch_size=number_of_structures(graph),
            )

        # generate prediction:
        if mode == "train":
            desired_properties = self.train_properties
        else:
            desired_properties = list(graph.properties.keys())

        predictions = self.model.predict(graph, properties=desired_properties)

        # compute the loss and its sub-components
        total_loss_result = self.total_loss(self.model, graph, predictions)

        # log
        log("loss/total", total_loss_result.loss_value)
        for name, loss_pair in total_loss_result.components.items():
            log(f"metrics/{name}", loss_pair.loss_value)
            log(f"loss/{name}_weighted", loss_pair.weighted_loss_value)

        # log additional values during validation
        if mode == "valid":
            for metric in self.eval_metrics:
                if metric.name in total_loss_result.components:
                    # don't double log
                    continue
                if any(
                    p not in graph.properties
                    for p in metric.required_properties
                ):
                    warnings.warn(
                        f"Metric {metric.name} requires properties "
                        f"{metric.required_properties} that are "
                        "not present in the graph: "
                        f"{list(graph.properties.keys())}. We won't log this "
                        "in this batch",
                        stacklevel=2,
                    )
                    continue
                value = metric(self.model, graph, predictions)
                log(f"metrics/{metric.name}", value)

        return total_loss_result.loss_value

    def training_step(self, structure: AtomicGraph, _):
        return self._step(structure, "train")

    def validation_step(self, structure: AtomicGraph, _):
        return self._step(structure, "valid")

    def configure_optimizers(
        self,
    ) -> torch.optim.Optimizer | OptimizerLRSchedulerConfig:
        opt = self.optimizer_factory(self.model)
        if self.scheduler_factory is None:
            return opt

        scheduler = self.scheduler_factory(opt)

        if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
            trainer = self.trainer
            assert trainer is not None

            # if the user is validating once every n training epochs,
            # set the scheduler to update every n training epochs
            if trainer.check_val_every_n_epoch:
                config = {
                    "interval": "epoch",  # really this means training epoch
                    "frequency": trainer.check_val_every_n_epoch,
                    "monitor": VALIDATION_LOSS_KEY,
                    "strict": True,
                }

            # otherwise, they are validating every n training steps:
            # set the scheduler to update every n training steps
            else:
                if isinstance(trainer.val_check_interval, float):
                    raise ValueError(
                        "`graph-pes` does not support specifying "
                        "`val_check_interval` as a fraction of the training "
                        "data size. Please specify it as an integer number of "
                        "training steps."
                    )
                config = {
                    "interval": "step",
                    "frequency": trainer.val_check_interval,
                    "monitor": VALIDATION_LOSS_KEY,
                    "strict": True,
                }

        else:
            # vanilla LR scheduler: update every training step
            config = {
                "interval": "step",
                "frequency": 1,
            }

        logger.debug(f"Using LR scheduler config:\n{config}")
        config["scheduler"] = scheduler
        return {"optimizer": opt, "lr_scheduler": config}  # type: ignore

    # we override the defaults to turn on gradient tracking for the
    # validation/test steps since we (might) compute the forces using autograd
    def on_validation_model_eval(self, *args, **kwargs):
        super().on_validation_model_eval(*args, **kwargs)
        torch.set_grad_enabled(True)

    @classmethod
    def load_best_weights(
        cls,
        model: GraphPESModel,
        trainer: pl.Trainer | None = None,
        checkpoint_path: Path | str | None = None,
    ):
        if checkpoint_path is None and trainer is None:
            raise ValueError(
                "Either trainer or checkpoint_path must be provided"
            )
        if checkpoint_path is None:
            path = trainer.checkpoint_callback.best_model_path  # type: ignore
        else:
            path = Path(checkpoint_path)
        logger.info(f'Loading best weights from "{path}"')
        checkpoint = torch.load(path, weights_only=True)
        state_dict = {
            k.replace("model.", "", 1): v
            for k, v in checkpoint["state_dict"].items()
            if k.startswith("model.")
        }
        model.load_state_dict(state_dict)


def test_with_lightning(
    trainer: pl.Trainer,
    model: GraphPESModel,
    data: dict[str, GraphDataset],
    loader_kwargs: dict,
    logging_prefix: str,
    user_eval_metrics: list[Loss] | None = None,
):
    assert trainer.test_loop.inference_mode is False

    test_loaders = [
        GraphDataLoader(dataset, **{**loader_kwargs, "shuffle": False})
        for dataset in data.values()
    ]

    task = TestingTask(
        model,
        list(data.keys()),
        get_all_eval_metrics(data.values(), user_eval_metrics),
        logging_prefix,
    )
    trainer.test(task, test_loaders)


class TestingTask(pl.LightningModule):
    def __init__(
        self,
        model: GraphPESModel,
        test_names: list[str],
        eval_metrics: list[Loss],
        logging_prefix: str,
    ):
        super().__init__()
        self.model = model
        self.test_names = test_names
        self.eval_metrics = eval_metrics
        self.test_properties = list(
            set.union(
                set(), *[set(m.required_properties) for m in eval_metrics]
            )
        )

    def test_step(
        self, structure: AtomicGraph, batch_idx: int, dataloader_idx: int = 0
    ):
        predictions = self.model.predict(
            structure, properties=self.test_properties
        )
        test_name = self.test_names[dataloader_idx]
        if test_name == "test" and len(self.test_names) == 1:
            prefix = "test"
        else:
            prefix = f"test/{test_name}"

        for metric in self.eval_metrics:
            if not all(
                p in structure.properties for p in metric.required_properties
            ):
                continue

            value = metric(self.model, structure, predictions)
            self.log(
                f"{prefix}/{metric.name}",
                value,
                on_epoch=True,
                sync_dist=True,
                batch_size=number_of_structures(structure),
                add_dataloader_idx=False,
            )

        return torch.tensor(0.0, requires_grad=True)

    def configure_optimizers(self):
        # no need for optimizer or scheduler during testing
        return None

    def on_test_model_eval(self):
        self.model.eval()
        # we need grad enabled to compute forces using autograd
        torch.set_grad_enabled(True)


def get_eval_metrics_for(dataset: GraphDataset) -> list[Loss]:
    evals = []

    if "energy" in dataset.properties:
        evals.append(PerAtomEnergyLoss(metric=RMSE()))
        evals.append(PerAtomEnergyLoss(metric=MAE()))
        evals.append(PropertyLoss("energy", RMSE()))
        evals.append(PropertyLoss("energy", MAE()))
    if "forces" in dataset.properties:
        # Force MAE is not invariant wrt. rotations, so we don't log it
        # see "How to validate machine-learned interatomic potentials"
        #      -> https://doi.org/10.1063/5.0139611
        evals.append(PropertyLoss("forces", RMSE()))
    if "stress" in dataset.properties:
        evals.append(PropertyLoss("stress", RMSE()))
    if "virial" in dataset.properties:
        evals.append(PropertyLoss("virial", RMSE()))

    return evals


def get_all_eval_metrics(
    data_sources: Iterable[GraphDataset],
    user_specified_metrics: list[Loss] | None = None,
) -> list[Loss]:
    if user_specified_metrics is None:
        user_specified_metrics = []

    # de-duplicate based on the name of the metric
    de_duped_metrics = {}
    for data_source in data_sources:
        de_duped_metrics.update(
            {m.name: m for m in get_eval_metrics_for(data_source)}
        )
    de_duped_metrics.update({m.name: m for m in user_specified_metrics})

    return list(de_duped_metrics.values())



from __future__ import annotations

import pathlib

import e3nn.util.jit
import torch

from graph_pes.atomic_graph import AtomicGraph, PropertyKey
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.misc import full_3x3_to_voigt_6


def as_lammps_data(
    graph: AtomicGraph,
    compute_virial: bool = False,
    debug: bool = False,
) -> dict[str, torch.Tensor]:
    return {
        "atomic_numbers": graph.Z,
        "positions": graph.R,
        "cell": graph.cell,
        "neighbour_list": graph.neighbour_list,
        "neighbour_cell_offsets": graph.neighbour_cell_offsets,
        "compute_virial": torch.tensor(compute_virial),
        "debug": torch.tensor(debug),
    }


class LAMMPSModel(torch.nn.Module):
    def __init__(self, model: GraphPESModel):
        super().__init__()
        self.model = model

    @torch.jit.export  # type: ignore
    def get_cutoff(self) -> torch.Tensor:
        return self.model.cutoff

    def forward(
        self, graph_data: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor]:
        debug = graph_data.get("debug", torch.tensor(False)).item()

        if debug:
            print("Received graph:")
            for key, value in graph_data.items():
                print(f"{key}: {value}")

        compute_virial = graph_data["compute_virial"].item()
        properties: list[PropertyKey] = ["energy", "forces", "local_energies"]
        if compute_virial:
            properties.append("virial")

        # graph_data is a dict, so we need to convert it to an AtomicGraph
        graph = AtomicGraph(
            Z=graph_data["atomic_numbers"],
            R=graph_data["positions"],
            cell=graph_data["cell"],
            neighbour_list=graph_data["neighbour_list"],
            neighbour_cell_offsets=graph_data["neighbour_cell_offsets"],
            properties={},
            other={},
            cutoff=self.model.cutoff.item(),
        )
        preds = self.model.predict(graph, properties=properties)

        # cast to float64
        for key in preds:
            preds[key] = preds[key].double()

        # add virial output if required
        if compute_virial:
            # LAMMPS expects the **virial** in Voigt notation
            # we provide the **stress** in full 3x3 matrix notation
            # therefore, convert:
            preds["virial"] = full_3x3_to_voigt_6(preds["virial"])

        return preds  # type: ignore

    def __call__(
        self, graph_data: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor]:
        return super().__call__(graph_data)


def deploy_model(model: GraphPESModel, path: str | pathlib.Path):
    """
    Deploy a :class:`~graph_pes.GraphPESModel` for use with LAMMPS.

    Use the resulting model with LAMMPS according to:

    .. code-block:: bash

        pair_style graph_pes <cpu>
        pair_coeff * * path/to/model.pt <element-of-type-1> <element-of-type-2> ...

    Parameters
    ----------
    model
        The model to deploy.
    path
        The path to save the deployed model to.
    """  # noqa: E501
    lammps_model = LAMMPSModel(model)
    scripted_model = e3nn.util.jit.script(lammps_model)
    torch.jit.save(scripted_model, path)



from __future__ import annotations

import time
import warnings
from typing import Iterable, TypeVar, overload

import ase
import numpy
import torch
from ase.calculators.calculator import Calculator, all_changes
from ase.stress import full_3x3_to_voigt_6_stress

from graph_pes.atomic_graph import AtomicGraph, PropertyKey, has_cell, to_batch
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.utils.misc import groups_of, pairs, uniform_repr


class GraphPESCalculator(Calculator):
    """
    ASE calculator wrapping any :class:`~graph_pes.GraphPESModel`.

    Implements a neighbour list caching scheme (see below) controlled by
    the ``skin`` parameter. Using ``skin > 0.0`` will

    * accelerate MD and minimisations
    * slow down single point calculations

    If you are predomintantly doing single point calculations, use
    ``skin=0``, otherwise, tune the ``skin`` paramter for your use case
    (see below).

    Parameters
    ----------
    model
        The model to wrap
    device
        The device to use for the calculation, e.g. "cpu" or "cuda".
        Defaults to ``None``, in which case the model is not moved
        from its current device.
    skin
        The additional skin to use for neighbour list calculations.
        If all atoms have moved less than half of this distance between
        calls to `calculate`, the neighbour list will be reused, saving
        (in some cases) significant computation time.
    **kwargs
        Properties passed to the :class:`ase.calculators.calculator.Calculator`
        base class.
    """

    implemented_properties = ["energy", "forces", "stress"]

    def __init__(
        self,
        model: GraphPESModel,
        device: torch.device | str | None = None,
        skin: float = 1.0,
        **kwargs,
    ):
        super().__init__(**kwargs)

        device = model.device if device is None else device
        self.model = model.to(device)
        self.model.eval()

        # caching for accelerated MD / calculation
        self._cached_graph: AtomicGraph | None = None
        self._cached_R: numpy.ndarray | None = None
        self._cached_cell: numpy.ndarray | None = None
        self.skin = skin

        # stats
        self.cache_hits = 0
        self.total_calls = 0
        self.nl_timings = []

    def calculate(
        self,
        atoms: ase.Atoms | None = None,
        properties: list[str] | list[PropertyKey] | None = None,
        system_changes: list[str] = all_changes,
    ):
        """
        Calculate the requested properties for the given structure, and store
        them to ``self.results``, as per a normal
        :class:`ase.calculators.calculator.Calculator`.

        Underneath-the-hood, this uses a neighbour list cache to speed up
        repeated calculations on the similar structures (i.e. particularly
        effective for MD and relaxations).
        """
        # handle defaults
        if properties is None:
            properties = ["energy", "forces"]

        # call to base-class to ensure setting of atoms attribute
        super().calculate(atoms, properties, system_changes)
        assert isinstance(self.atoms, ase.Atoms)

        self.total_calls += 1

        graph: AtomicGraph | None = None

        # avoid re-calculating neighbour lists if possible
        if (
            set(system_changes) <= {"positions", "cell"}
            and self._cached_graph is not None
            and self._cached_R is not None
            and self._cached_cell is not None
        ):
            new_R = self.atoms.positions
            new_cell = self.atoms.cell.array
            changes = numpy.linalg.norm(new_R - self._cached_R, axis=-1)
            cell_changes = numpy.linalg.norm(
                new_cell - self._cached_cell, axis=-1
            )
            # cache hit
            if numpy.all(changes < self.skin / 2) and numpy.all(
                cell_changes < self.skin / 2
            ):
                self.cache_hits += 1
                graph = self._cached_graph._replace(
                    R=torch.tensor(new_R, dtype=self._cached_graph.R.dtype),
                    cell=torch.tensor(
                        new_cell, dtype=self._cached_graph.cell.dtype
                    ),
                )

        # cache miss
        if graph is None:
            tick = time.perf_counter()
            graph = AtomicGraph.from_ase(
                self.atoms, self.model.cutoff.item() + self.skin
            ).to(self.model.device)
            tock = time.perf_counter()
            self.nl_timings.append(tock - tick)
            self._cached_graph = graph
            self._cached_R = graph.R.detach().cpu().numpy()
            self._cached_cell = graph.cell.detach().cpu().numpy()

        graph = graph.to(self.model.device)

        results = {
            k: v.detach().cpu().numpy()
            for k, v in self.model.predict(
                graph,
                properties=properties,  # type: ignore
            ).items()
            if k in properties
        }
        if "energy" in properties:
            results["energy"] = results["energy"].item()
        if "stress" in properties:
            results["stress"] = full_3x3_to_voigt_6_stress(results["stress"])

        self.results = results

    @property
    def cache_hit_rate(self) -> float:
        """
        The ratio of calls to
        :meth:`~graph_pes.utils.calculator.GraphPESCalculator.calculate`
        for which the neighbour list was reused.
        """
        if self.total_calls == 0:
            warnings.warn("No calls to calculate yet", stacklevel=2)
            return 0.0
        return self.cache_hits / self.total_calls

    @property
    def average_nl_timing(self) -> float:
        """The average time taken to calculate the neighbour list in seconds."""
        return numpy.mean(self.nl_timings).item()

    @property
    def total_nl_timing(self) -> float:
        """The total time taken to calculate the neighbour list in seconds."""
        return sum(self.nl_timings)

    def reset_cache_stats(self):
        """Reset the :attr:`cache_hit_rate` statistic."""
        self.cache_hits = 0
        self.total_calls = 0
        self.nl_timings = []

    def calculate_all(
        self,
        structures: Iterable[AtomicGraph | ase.Atoms],
        properties: list[PropertyKey] | None = None,
        batch_size: int = 5,
    ) -> list[dict[PropertyKey, numpy.ndarray | float]]:
        """
        Semantically identical to:

        .. code-block::

            [calc.calculate(structure, properties) for structure in structures]

        but with significant acceleration due to internal batching.

        Parameters
        ----------
        structures
            A list of :class:`~graph_pes.AtomicGraph` or
            :class:`ase.Atoms` objects.
        properties
            The properties to predict.
        batch_size
            The number of structures to predict at once.
        """
        # Convert structures to graphs and handle defaults
        graphs = [
            AtomicGraph.from_ase(s, self.model.cutoff.item() + 0.001)
            if isinstance(s, ase.Atoms)
            else s
            for s in structures
        ]
        graphs = [g.to(self.model.device) for g in graphs]

        if properties is None:
            properties = ["energy", "forces"]
            if all(map(has_cell, graphs)):
                properties.extend(["stress", "virial"])

        # Batched prediction
        tensor_results: list[dict[PropertyKey, torch.Tensor]] = []
        for batch in map(to_batch, groups_of(batch_size, graphs)):
            predictions = self.model.predict(batch, properties)
            separated = _seperate(predictions, batch)
            tensor_results.extend(separated)

        results: list[dict[PropertyKey, numpy.ndarray | float]] = [
            {k: to_numpy(v) for k, v in r.items()} for r in tensor_results
        ]

        # Convert stress tensors to Voigt notation
        for r in results:
            for key in ["stress", "virial"]:
                if key in r:
                    r[key] = full_3x3_to_voigt_6_stress(r[key])

        return results

    def __repr__(self):
        return uniform_repr(
            self.__class__.__name__,
            model=self.model,
            device=self.model.device,
            skin=self.skin,
        )


## utils ##

T = TypeVar("T")
TensorLike = TypeVar("TensorLike", torch.Tensor, numpy.ndarray)


def to_numpy(t: torch.Tensor) -> numpy.ndarray | float:
    n = t.detach().cpu().numpy()
    return n.item() if n.shape == () else n


def _seperate(
    batched_prediction: dict[PropertyKey, TensorLike],
    batch: AtomicGraph,
) -> list[dict[PropertyKey, TensorLike]]:
    preds_list = []
    assert batch.ptr is not None

    for idx, (start, stop) in enumerate(pairs(batch.ptr)):
        preds = {}

        # per-structure properties
        if "energy" in batched_prediction:
            preds["energy"] = batched_prediction["energy"][idx]
        if "stress" in batched_prediction:
            preds["stress"] = batched_prediction["stress"][idx]
        if "virial" in batched_prediction:
            preds["virial"] = batched_prediction["virial"][idx]

        # per-atom properties
        if "forces" in batched_prediction:
            preds["forces"] = batched_prediction["forces"][start:stop]
        if "local_energies" in batched_prediction:
            preds["local_energies"] = batched_prediction["local_energies"][
                start:stop
            ]

        preds_list.append(preds)

    return preds_list


@overload
def merge_predictions(  # type: ignore
    predictions: list[dict[PropertyKey, numpy.ndarray | float]],
) -> dict[PropertyKey, numpy.ndarray]: ...
@overload
def merge_predictions(
    predictions: list[dict[PropertyKey, torch.Tensor | float]],
) -> dict[PropertyKey, torch.Tensor]: ...
def merge_predictions(
    predictions: list[dict[PropertyKey, TensorLike | float]],
) -> dict[PropertyKey, TensorLike]:
    """
    Take a list of property predictions and merge them
    in a sensible way. Implemented for both :class:`torch.Tensor`
    and :class:`numpy.ndarray`.

    Parameters
    ----------
    predictions
        A list of property predictions each corresponding to a single
        structure.

    Examples
    --------
    >>> predictions = [
    ...     {"energy": np.array(1.0), "forces": np.array([[1, 2], [3, 4]])},
    ...     {"energy": np.array(2.0), "forces": np.array([[5, 6], [7, 8]])},
    ... ]
    >>> merge_predictions(predictions)
    {'energy': array([1., 2.]), 'forces': array([[1, 2], [3, 4], [5, 6], [7, 8]])}
    """  # noqa: E501
    if not predictions:
        return {}

    eg = next(iter(predictions[0].values()))
    if isinstance(eg, torch.Tensor):
        stack = torch.stack
        cat = torch.cat
    else:
        stack = numpy.stack
        cat = numpy.concatenate

    merged: dict[PropertyKey, TensorLike] = {}

    # stack per-structure properties along new axis
    for key in ["energy", "stress"]:
        if key in predictions[0]:
            merged[key] = stack([p[key] for p in predictions])  # type: ignore

    # concatenat per-atom properties along the first axis
    for key in ["forces", "local_energies"]:
        if key in predictions[0]:
            merged[key] = cat([p[key] for p in predictions])  # type: ignore

    return merged



from __future__ import annotations

import torch

from graph_pes.atomic_graph import (
    AtomicGraph,
    neighbour_distances,
    neighbour_vectors,
    number_of_atoms,
    number_of_edges,
)


def angle_spanned_by(v1: torch.Tensor, v2: torch.Tensor):
    """
    Calculate angles between corresponding vectors in two batches.

    Parameters
    ----------
    v1
        First batch of vectors, shape (N, 3)
    v2
        Second batch of vectors, shape (N, 3)

    Returns
    -------
    torch.Tensor
        Angles in radians, shape (N,)
    """
    # Compute dot product
    dot_product = torch.sum(v1 * v2, dim=1)

    # Compute magnitudes
    v1_mag = torch.linalg.vector_norm(v1, dim=1)
    v2_mag = torch.linalg.vector_norm(v2, dim=1)

    # Compute cosine of angle, add small epsilon to prevent division by zero
    cos_angle = dot_product / (v1_mag * v2_mag)

    # Clamp cosine values to handle numerical instabilities
    cos_angle = torch.clamp(cos_angle, min=-1.0 + 1e-7, max=1.0 - 1e-7)

    # Compute angle using arccos
    return torch.arccos(cos_angle)


def triplet_bond_descriptors(
    graph: AtomicGraph,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    For each triplet :math:`(i, j, k)`, get the bond angle :math:`\theta_{jik}`
    (in radians) and the two bond lengths :math:`r_{ij}` and :math:`r_{ik}`.

    Returns
    -------
    triplet_idxs
        The triplet indices, :math:`(i, j, k)`, of shape ``(Y, 3)``.
    angle
        The bond angle :math:`\theta_{jik}`, shape ``(Y,)``.
    r_ij
        The bond length :math:`r_{ij}`, shape ``(Y,)``.
    r_ik
        The bond length :math:`r_{ik}`, shape ``(Y,)``.

    Examples
    --------
    >>> graph = AtomicGraph.from_ase(molecule("H2O"))
    >>> angle, r_ij, r_ik = triplet_bond_descriptors(graph)
    >>> torch.rad2deg(angle)
    tensor([103.9999, 103.9999,  38.0001,  38.0001,  38.0001,  38.0001])
    """

    edge_pairs = triplet_edge_pairs(graph, graph.cutoff)  # (Y, 2)

    ij = graph.neighbour_list[:, edge_pairs[:, 0]]  # (2, Y)
    k = graph.neighbour_list[1, edge_pairs[:, 1]].unsqueeze(0)  # (1, Y)
    triplet_idxs = torch.cat([ij, k], dim=0).transpose(0, 1)  # (Y, 3)

    if triplet_idxs.shape[0] == 0:
        return (
            triplet_idxs,
            torch.zeros(0, device=graph.R.device).float(),
            torch.zeros(0, device=graph.R.device).float(),
            torch.zeros(0, device=graph.R.device).float(),
        )

    v = neighbour_vectors(graph)
    v1 = v[edge_pairs[:, 0]]
    v2 = v[edge_pairs[:, 1]]

    return (
        triplet_idxs,
        angle_spanned_by(v1, v2),
        torch.linalg.vector_norm(v1, dim=-1),
        torch.linalg.vector_norm(v2, dim=-1),
    )


@torch.no_grad()
def triplet_edge_pairs(graph: AtomicGraph, three_body_cutoff: float):
    r"""
    Find all the pairs of edges, :math:`a = (i, j), b = (i, k)`, such that:

    * :math:`i, j, k \in \{0, 1, \dots, N-1\}` are indices of distinct
      (images of) atoms within the graph
    * :math:`j \neq k`
    * :math:`r_{ij} \leq` ``three_body_cutoff``
    * :math:`r_{ik} \leq` ``three_body_cutoff``

    Returns
    -------
    edge_pairs: torch.Tensor
        A ``(Y, 2)`` shaped tensor indicating the edges, such that

        .. code-block:: python

            a, b = edge_pairs[y]
            i, j = graph.neighbour_list[:,a]
            i, k = graph.neighbour_list[:,b]
    """

    edge_indexes = torch.arange(number_of_edges(graph), device=graph.R.device)

    three_body_mask = neighbour_distances(graph) < three_body_cutoff
    relevant_edge_indexes = edge_indexes[three_body_mask]
    relevant_central_atoms = graph.neighbour_list[0][relevant_edge_indexes]

    edge_pairs = []

    for i in range(number_of_atoms(graph)):
        mask = relevant_central_atoms == i
        masked_edge_indexes = relevant_edge_indexes[mask]

        # number of edges of distance <= three_body_cutoff
        # that have i as a central atom
        N = masked_edge_indexes.shape[0]
        _idx = torch.cartesian_prod(
            torch.arange(N),
            torch.arange(N),
        )  # (N**2, 2)
        _idx = _idx[_idx[:, 0] != _idx[:, 1]]  # (N**2 - N, 2)

        pairs_for_i = masked_edge_indexes[_idx]
        edge_pairs.append(pairs_for_i)

    edge_pairs = torch.cat(edge_pairs)

    return edge_pairs



import warnings
from typing import (
    TYPE_CHECKING,
    Dict,
    Final,
    List,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Protocol,
    Sequence,
    Union,
    cast,
)

import ase
import numpy as np
import torch
import torch.multiprocessing
import torch.utils.data
import vesin
from ase.stress import voigt_6_to_full_3x3_stress
from load_atoms.utils import remove_calculator
from typing_extensions import TypeAlias

from graph_pes.utils.misc import (
    all_equal,
    is_being_documented,
    left_aligned_div,
    left_aligned_mul,
    to_significant_figures,
    uniform_repr,
)

DEFAULT_CUTOFF: Final[float] = 5.0


PropertyKey: TypeAlias = Literal[
    "local_energies", "forces", "energy", "stress", "virial"
]
ALL_PROPERTY_KEYS: Final[List[PropertyKey]] = [
    "local_energies",
    "forces",
    "energy",
    "stress",
    "virial",
]

if not TYPE_CHECKING and not is_being_documented():
    # torchscript doesn't handle TypedDicts or Literal types:
    # at run-time, we just use less specific, but still correct, types
    Properties: TypeAlias = Dict[str, torch.Tensor]
    PropertyKey: TypeAlias = str


class AtomicGraph(NamedTuple):
    r"""
    An :class:`AtomicGraph` represents an atomic structure.
    Each node corresponds to an atom, and each directed edge links a central
    atom to a "bonded" neighbour.

    We implement such graphs as (immutable) :class:`~typing.NamedTuple`\ s.
    This allows for easy serialisation and compatibility with PyTorch,
    TorchScript and other libraries. These objects, and all functions that
    operate on them, are compatible with both isolated and periodic structures.

    Batches of multiple structures are represented by a single
    :class:`AtomicGraph` object containing multiple disjoint subgraphs. See
    :func:`~graph_pes.atomic_graph.to_batch` for more information.

    Properties
    ++++++++++

    Below we assume the graph contains ``N`` atoms, ``E`` edges (and ``S``
    structures if the graph is batched).

    We use two examples to illustrate the various properties:


    .. dropdown:: Water molecule

        .. code-block:: python

            >>> from ase.build import molecule
            >>> from load_atoms import view
            >>> water = molecule("H2O")
            >>> view(water, show_bonds=True)

        .. raw:: html
            :file: ../../docs/source/_static/water.html

        We can see that there are 3 atoms, with 2 bonds (and therefore 4
        directed edges):

        .. code-block:: python

            >>> from graph_pes import AtomicGraph
            >>> water_graph = AtomicGraph.from_ase(water, cutoff=1.2)
            >>> water_graph
            AtomicGraph(atoms=3, edges=4, has_cell=False, cutoff=1.2)

    .. dropdown:: Sodium crystal

        .. code-block:: python

            >>> from ase.build import bulk
            >>> from load_atoms import view
            >>> sodium = bulk("Na")
            >>> view(sodium.repeat(3), show_bonds=True)

        .. raw:: html
            :file: ../../docs/source/_static/Na.html

        This structure has a single atom within a periodic cell. If you
        look closely, you can see that this atom has 8 nearest neighbours.
        Only "source" atoms within the unit cell are included in the
        neighbour list, and hence there are 8 edges:

        .. code-block:: python

            >>> sodium_graph = AtomicGraph.from_ase(sodium, cutoff=3.7)
            >>> sodium_graph
            AtomicGraph(atoms=1, edges=8, has_cell=True, cutoff=3.7)
    """

    Z: torch.Tensor
    """
    The atomic numbers of the atoms in the graph, of shape ``(N,)``.

    .. code-block:: python

        >>> water_graph.Z
        tensor([8, 1, 1])
    
    .. code-block:: python

        >>> sodium_graph.Z
        tensor([11])
    """

    R: torch.Tensor
    """
    The cartesian positions of the atoms in the graph, of shape ``(N, 3)``.

    .. code-block:: python

        >>> water_graph.R
        tensor([[ 0.0000,  0.0000,  0.1193],
                [ 0.0000,  0.7632, -0.4770],
                [ 0.0000, -0.7632, -0.4770]])
    
    .. code-block:: python
        
        >>> sodium_graph.R
        tensor([[0., 0., 0.]])
    """

    cell: torch.Tensor
    """
    The unit cell vectors, of shape ``(3, 3)`` for a single structure, or
    ``(S, 3, 3)`` for a batched graph. If the structure is non-periodic,
    this will be all zeros.

    .. code-block:: python

        >>> water_graph.cell
        tensor([[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]])
    
    .. code-block:: python
        
        >>> sodium_graph.cell
        tensor([[-2.1150,  2.1150,  2.1150],
                [ 2.1150, -2.1150,  2.1150],
                [ 2.1150,  2.1150, -2.1150]])
    """

    neighbour_list: torch.Tensor
    """
    A neighbour list, of shape ``(2, E)``, where 
    ``i, j = graph.neighbour_list[:, k]`` is the ``k``'th directed edge
    in the graph, linking atom ``i`` to atom ``j``.

    .. code-block:: python

        >>> water_graph.neighbour_list
        tensor([[0, 0, 1, 2],
                [1, 2, 0, 0]])
    
    .. code-block:: python
        
        >>> sodium_graph.neighbour_list
        tensor([[0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0]])
    """

    neighbour_cell_offsets: torch.Tensor
    """
    The offsets of the neighbours of each atom in units of the unit cell
    vectors, of shape ``(E, 3)``, such that:

    .. code-block:: python

        # k-th edge
        i, j = graph.neighbour_list[:, k]
        kth_displacement_vector = (
            graph.R[j] 
            + graph.neighbour_cell_offsets[k] @ graph.cell
            - graph.R[i]
        )

    In the case of an isolated, non-periodic structure, these will be all zeros.

    .. code-block:: python

        >>> water_graph.neighbour_cell_offsets
        tensor([[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]])
    
    .. code-block:: python

        >>> sodium_graph.neighbour_cell_offsets
        tensor([[ 0.,  0.,  1.],
                [ 0.,  1.,  0.],
                [ 1.,  1.,  1.],
                [ 1.,  0.,  0.],
                [-1.,  0.,  0.],
                [-1., -1., -1.],
                [ 0., -1.,  0.],
                [ 0.,  0., -1.]])

    """

    properties: Dict[PropertyKey, torch.Tensor]
    """
    A dictionary containing potential energy surface (PES) related properties
    of the graph.

    .. list-table::
        :header-rows: 1

        * - Key
          - Shape
          - Description
        * - :code:`"local_energies"`
          - :code:`(N,)`
          - contribution to the total energy from each atom
        * - :code:`"energy"`
          - ``()`` 
          
            (``(S,)`` if batched)
          - total energy of the structure
        * - :code:`"forces"`
          - :code:`(N, 3)`
          - force on each atom
        * - :code:`"stress"`
          - ``(3, 3)`` 
          
            (``(S, 3, 3)`` if batched)
          - stress tensor (see :doc:`../theory`)
        * - :code:`"virial"`
          - ``(3, 3)`` 
          
            (``(S, 3, 3)`` if batched)
          - virial stress tensor (see :doc:`../theory`)
    """

    cutoff: float
    """
    The cutoff distance used to create the neighbour list for this graph.
    """

    other: Dict[str, torch.Tensor]
    """
    A dictionary containing any other additional information about the graph.
    Feel free to populate this as you wish.
    """

    batch: Union[torch.Tensor, None] = None
    """
    A tensor of shape ``(N,)`` indicating the index ``i`` of the structure
    within the batch that each atom belongs to. Not present for a single
    structure.
    """

    ptr: Union[torch.Tensor, None] = None
    """
    A tensor of shape ``(S + 1,)`` indicating the index ``i`` of the first
    atom in each structure within the batch. Not present for a single structure.
    """

    pbc: Union[torch.Tensor, None] = None
    """
    Whether the structure has periodic boundary conditions (i.e., is periodic)
    in each of the three spatial dimensions: ``(x, y, z)``.  Shape ``(3,)``.
    DType: :code:`torch.bool`.

    ``graph-pes`` does not use this information directly (and hence this field
    is options), but it can be useful for external tools to have access to this,
    e.g. when converting to and from an :class:`ase.Atoms` object.

    .. code-block:: python

        >>> assert atoms.pbc
        >>> graph = AtomicGraph.from_ase(atoms)
        >>> atoms = graph.to_ase()
        >>> atoms.pbc
        tensor([ True,  True,  True])
    """

    @classmethod
    def from_ase(
        cls,
        structure: ase.Atoms,
        cutoff: float = DEFAULT_CUTOFF,
        property_mapping: Union[Mapping[str, PropertyKey], None] = None,
        others_to_include: Union[Sequence[str], None] = None,
    ) -> "AtomicGraph":
        r"""
        Convert an :class:`ase.Atoms` object to an :class:`AtomicGraph`.

        Parameters
        ----------
        structure
            The :class:`ase.Atoms` object to convert.
        cutoff
            The cutoff distance for neighbour finding.
        property_mapping
            An optional mapping of the form ``{key_on_structure:
            key_for_graph}`` defining how relevant properties are labelled on
            the :class:`ase.Atoms` object. If not provided, this function will
            extract all of ``"energy"``, ``"forces"``, ``"stress"``, or
            ``"virial"`` from the ``.info`` and ``.arrays`` dicts if they are
            present.
        others_to_include
            An optional list of other ``.info``/``.arrays`` keys to include in
            the graph's ``other`` dict. The corresponding values will be
            converted to :class:`torch.Tensor`\ s.

        Example
        -------

        .. code-block:: python

            >>> from ase.build import molecule
            >>> from graph_pes import AtomicGraph

            >>> # create a a structure with some extra info
            >>> atoms = molecule("H2O")
            >>> atoms.info["DFT_energy"] = -10.0
            >>> atoms.info["unique_id"] = 1234

            >>> # default behaviour:
            >>> AtomicGraph.from_ase(atoms)
            AtomicGraph(atoms=3, edges=6, has_cell=False, cutoff=5.0)

            >>> # specify how to map properties, and other things to include
            >>> AtomicGraph.from_ase(
            ...     atoms,
            ...     property_mapping={
            ...         "DFT_energy": "energy",
            ...     },
            ...     others_to_include=["unique_id"],
            ... )
            AtomicGraph(
                atoms=3,
                edges=6,
                has_cell=False,
                cutoff=5.0,
                properties=['energy'],
                other=['unique_id']
            )
        """

        # account for strange behaviour in ase 3.23.0+ whereby
        # properties are sometimes removed from the atoms.info/arrays dicts
        # to ensure we don't change the atoms object that the user passes in
        # we first make a copy, ensuring that the calculator is also copied over
        calc = structure.calc
        structure = structure.copy()
        structure.calc = calc
        # then we remove the calculator and copy over the properties to the
        # relevant info/arrays dicts (with help from load_atoms.utils)
        remove_calculator(structure)

        _float = torch.get_default_dtype()

        # structure
        Z = torch.tensor(structure.numbers, dtype=torch.long)
        R = torch.tensor(structure.positions, dtype=_float)
        cell = torch.tensor(structure.cell.array, dtype=_float)

        # neighbour list
        if structure.pbc.any() and np.all(structure.cell == 0):
            raise ValueError(
                "PBCs are set to True, but cell is all zeros: we can't "
                "create a neighbour list. Please set the cell or switch off "
                "PBCs."
            )
        i, j, offsets = vesin.ase_neighbor_list("ijS", structure, float(cutoff))
        i = i.astype(np.int64)
        j = j.astype(np.int64)
        neighbour_list = torch.tensor(np.vstack([i, j]), dtype=torch.long)
        neighbour_cell_offsets = torch.tensor(offsets, dtype=_float)

        # properties
        properties: dict[PropertyKey, torch.Tensor] = {}
        other: dict[str, torch.Tensor] = {}

        if property_mapping is None:
            all_keys = set(structure.info) | set(structure.arrays)
            property_mapping = {
                k: cast(PropertyKey, k)
                for k in ["energy", "forces", "stress", "virial"]
                if k in all_keys
            }
        if others_to_include is None:
            others_to_include = []

        def to_tensor(value):
            t = torch.tensor(value)
            if t.is_floating_point():
                t = t.to(_float)
            return t

        for key, value in list(structure.info.items()) + list(
            structure.arrays.items()
        ):
            if key in property_mapping:
                property = property_mapping[key]
                # ensure stress is always 3x3, not voigt notation
                if property in ["stress", "virial"] and value.reshape(
                    -1
                ).shape == (6,):
                    value = voigt_6_to_full_3x3_stress(value)
                properties[property] = to_tensor(value)

            elif key in others_to_include:
                other[key] = to_tensor(value)

        missing = set(
            structure_key
            for structure_key, graph_key in property_mapping.items()
            if graph_key not in properties
        )
        if missing:
            raise ValueError(f"Unable to find properties: {missing}")

        pbc = torch.tensor(structure.pbc, dtype=torch.bool)

        return cls(
            Z=Z,
            R=R,
            cell=cell,
            neighbour_list=neighbour_list,
            neighbour_cell_offsets=neighbour_cell_offsets,
            properties=properties,
            other=other,
            cutoff=cutoff,
            pbc=pbc,
        )

    @classmethod
    def create_with_defaults(
        cls,
        Z: torch.Tensor,
        R: torch.Tensor,
        cell: Union[torch.Tensor, None] = None,
        neighbour_list: Union[torch.Tensor, None] = None,
        neighbour_cell_offsets: Union[torch.Tensor, None] = None,
        properties: Union[Dict[PropertyKey, torch.Tensor], None] = None,
        other: Union[Dict[str, torch.Tensor], None] = None,
        cutoff: float = 0.0,
        pbc: Union[torch.Tensor, None] = None,
    ) -> "AtomicGraph":
        """
        Create an :class:`AtomicGraph`, populating missing values with defaults.

        Parameters
        ----------
        Z
            The atomic numbers.
        R
            The cartesian positions.
        cell
            The unit cell. Defaults to ``torch.zeros(3, 3)``.
        neighbour_list
            The neighbour list. Defaults to ``torch.zeros(2, 0)``.
        neighbour_cell_offsets
            The neighbour cell offsets. Defaults to ``torch.zeros(0, 3)``.
        properties
            The properties. Defaults to ``{}``.
        other
            The other information. Defaults to ``{}``.
        """

        if cell is None:
            cell = torch.zeros(3, 3, device=R.device).float()
        if neighbour_list is None:
            neighbour_list = torch.zeros(2, 0, device=R.device).long()
        if neighbour_cell_offsets is None:
            neighbour_cell_offsets = torch.zeros(0, 3, device=R.device).float()
        if properties is None:
            properties = {}
        if other is None:
            other = {}
        return cls(
            Z=Z,
            R=R,
            cell=cell,
            neighbour_list=neighbour_list,
            neighbour_cell_offsets=neighbour_cell_offsets,
            properties=properties,
            other=other,
            cutoff=cutoff,
            pbc=pbc,
        )

    def to(self, device: Union[torch.device, str]) -> "AtomicGraph":
        """Move this graph to the specified device."""

        properties: dict[PropertyKey, torch.Tensor] = {
            k: v.to(device) for k, v in self.properties.items()
        }
        return AtomicGraph(
            Z=self.Z.to(device),
            R=self.R.to(device),
            cell=self.cell.to(device),
            neighbour_list=self.neighbour_list.to(device),
            neighbour_cell_offsets=self.neighbour_cell_offsets.to(device),
            properties=properties,
            other={k: v.to(device) for k, v in self.other.items()},
            cutoff=self.cutoff,
            batch=self.batch.to(device) if self.batch is not None else None,
            ptr=self.ptr.to(device) if self.ptr is not None else None,
            pbc=self.pbc.to(device) if self.pbc is not None else None,
        )

    def __repr__(self):
        info = {}

        if self.batch is not None:
            name = "AtomicGraphBatch"
            info["structures"] = self.batch.max().item() + 1
        else:
            name = "AtomicGraph"

        info["atoms"] = number_of_atoms(self)
        info["edges"] = number_of_edges(self)
        info["has_cell"] = has_cell(self)
        info["cutoff"] = to_significant_figures(self.cutoff, 3)
        if self.properties:
            info["properties"] = available_properties(self)
        if self.other:
            info["other"] = list(self.other.keys())

        return uniform_repr(name, **info, indent_width=4)

    def to_ase(self) -> ase.Atoms:
        """Convert this graph to an :class:`ase.Atoms` object."""

        pbc = self.pbc.cpu().numpy() if self.pbc is not None else None
        atoms = ase.Atoms(
            numbers=self.Z.detach().cpu().numpy(),
            positions=self.R.detach().cpu().numpy(),
            cell=self.cell.detach().cpu().numpy(),
            pbc=pbc,
        )
        for key, value in self.other.items():
            atoms.info[key] = value.detach().cpu().numpy()

        if "energy" in self.properties:
            atoms.info["energy"] = (
                self.properties["energy"].detach().cpu().item()
            )

        for key in ["stress", "virial"]:
            if key in self.properties:
                atoms.info[key] = self.properties[key].detach().cpu().numpy()

        for key in ["forces", "local_energies"]:
            if key in self.properties:
                atoms.arrays[key] = self.properties[key].detach().cpu().numpy()

        return atoms


def replace(
    graph: AtomicGraph,
    Z: Optional[torch.Tensor] = None,
    R: Optional[torch.Tensor] = None,
    cell: Optional[torch.Tensor] = None,
    neighbour_list: Optional[torch.Tensor] = None,
    neighbour_cell_offsets: Optional[torch.Tensor] = None,
    properties: Optional[dict[PropertyKey, torch.Tensor]] = None,
    other: Optional[dict[str, torch.Tensor]] = None,
    cutoff: Optional[float] = None,
    pbc: Optional[torch.Tensor] = None,
) -> AtomicGraph:
    """
    A convenience function for replacing the values of an :class:`AtomicGraph`
    that is ``TorchScript`` compatible (as opposed to the built-in ``._replace``
    namedtuple method).
    """
    return AtomicGraph(
        Z=Z if Z is not None else graph.Z,
        R=R if R is not None else graph.R,
        cell=cell if cell is not None else graph.cell,
        neighbour_list=neighbour_list
        if neighbour_list is not None
        else graph.neighbour_list,
        neighbour_cell_offsets=neighbour_cell_offsets
        if neighbour_cell_offsets is not None
        else graph.neighbour_cell_offsets,
        properties=properties if properties is not None else graph.properties,
        other=other if other is not None else graph.other,
        cutoff=cutoff if cutoff is not None else graph.cutoff,
        batch=graph.batch,
        ptr=graph.ptr,
        pbc=pbc if pbc is not None else graph.pbc,
    )


############################### BATCHING ###############################


class CustomPropertyBatcher(Protocol):
    def __call__(
        self, batch: AtomicGraph, values: list[torch.Tensor]
    ) -> torch.Tensor:
        """
        Batch the given values.

        Parameters
        ----------
        batch
            The batch of graphs.
        values
            The list of values to batch.
        """
        ...


_custom_batchers: dict[str, CustomPropertyBatcher] = {}

# NB this is essential, otherwise all data loader workers will
# have an empty _custom_batchers dict, and hence fail to perform
# any custom batching
torch.multiprocessing.set_start_method("fork", force=True)


def register_custom_batcher(key: str):
    """
    Register a custom batcher for a property in the ``other`` field.

    The batcher should conform to the following protocol:

    .. autoclass:: graph_pes.atomic_graph.CustomPropertyBatcher()
        :members: __call__

    Parameters
    ----------
    key
        The key of the property to register a custom batcher for.

    Examples
    --------

    >>> from graph_pes.atomic_graph import register_custom_batcher
    >>> @register_custom_batcher("foo")
    ... def foo_batcher(batch, values):
    ...     return torch.max(torch.vstack(values), dim=0).values
    >>> ... # create graphs
    >>> graphs[0].other["foo"], graphs[1].other["foo"]
    (tensor([1]), tensor([2]))
    >>> ... # batch the graphs
    >>> batch = to_batch(graphs)
    >>> batch.other["foo"]
    tensor([2])
    """

    def decorator(func: CustomPropertyBatcher):
        _custom_batchers[key] = func
        return func

    return decorator


def to_batch(
    graphs: Sequence[AtomicGraph],
) -> AtomicGraph:
    """
    Collate a sequence of atomic graphs into a single batch object.

    The ``Z``, ``R``, ``neighbour_list``, and ``neighbour_cell_offsets``
    properties are concatenated along the first axis, while the ``cell``
    property is stacked along a new batch dimension.

    Values in the ``"other"`` dictionary are concatenated along the first axis
    if they appear to be a per-atoms property (i.e. their first dimension
    matches the number of atoms in the structure). Otherwise, they are stacked
    along a new batch dimension.

    Parameters
    ----------
    graphs
        The graphs to collate.

    Examples
    --------

    A basic example:

    >>> from ase.build import molecule
    >>> from graph_pes import AtomicGraph, to_batch
    >>> graphs = [
    ...     AtomicGraph.from_ase(molecule("H2O")),
    ...     AtomicGraph.from_ase(molecule("CH4")),
    ... ]
    >>> batch = to_batch(graphs)
    >>> batch.batch  # H20 has 3 atoms, CH4 has 5
    tensor([0, 0, 0, 1, 1, 1, 1, 1])
    >>> batch.ptr  # offset of first atom of each graph
    tensor([0, 3, 8])
    >>> batch.Z.shape
    torch.Size([8])
    >>> batch.R.shape
    torch.Size([8, 3])
    >>> batch.cell.shape
    torch.Size([2, 3, 3])
    """
    if any(is_batch(g) for g in graphs):
        raise ValueError("Cannot recursively batch graphs")

    # easy properties: just cat these together
    Z = torch.cat([g.Z for g in graphs])
    R = torch.cat([g.R for g in graphs])
    neighbour_offsets = torch.cat([g.neighbour_cell_offsets for g in graphs])

    # stack cells along a new batch dimension
    if not all_equal([has_cell(g) for g in graphs]):
        warnings.warn(
            "Attempting to batch a colleciton of graphs where only some "
            "have a defined unit cell. This may lead to unexpected results.",
            stacklevel=2,
        )
    cells = torch.stack([g.cell for g in graphs])

    # standard way to caculaute the batch and ptr properties
    batch = torch.cat(
        [torch.full_like(g.Z, fill_value=i) for i, g in enumerate(graphs)]
    )
    ptr = torch.tensor([0] + [g.Z.shape[0] for g in graphs]).cumsum(dim=0)

    # use the ptr to increment the neighbour index appropriately
    neighbour_list = torch.cat(
        [g.neighbour_list + ptr[i] for i, g in enumerate(graphs)], dim=1
    )

    # handle cutoff
    cutoffs = [g.cutoff for g in graphs]
    if not all_equal(cutoffs):
        warnings.warn(
            "Attempting to batch graphs with different cutoffs: "
            f"{cutoffs}. Setting graph.cutoff to the maximum.",
            stacklevel=2,
        )
    cutoff = max(cutoffs)

    properties: dict[PropertyKey, torch.Tensor] = {}
    # - per structure labels are concatenated along a new batch axis (0)
    for key in ["energy", "stress", "virial"]:
        key = cast(PropertyKey, key)
        if all(key in g.properties for g in graphs):
            properties[key] = torch.stack([g.properties[key] for g in graphs])

    # - per atom labels are concatenated along the first axis
    for key in ["forces", "local_energies"]:
        key = cast(PropertyKey, key)
        if all(key in g.properties for g in graphs):
            properties[key] = torch.cat([g.properties[key] for g in graphs])

    batched_graph = AtomicGraph(
        Z=Z,
        R=R,
        cell=cells,
        neighbour_list=neighbour_list,
        neighbour_cell_offsets=neighbour_offsets,
        properties=properties,
        other={},
        cutoff=cutoff,
        batch=batch,
        ptr=ptr,
    )

    # - finally, add in the other stuff: this is a bit tricky
    #   since we need to try and infer whether these are per-atom
    #   or per-structure
    for key in graphs[0].other:
        values = [g.other[key] for g in graphs]
        if key in _custom_batchers:
            batcher = _custom_batchers[key]
            batched_graph.other[key] = batcher(batched_graph, values)
        elif all(is_local_property(g.other[key], g) for g in graphs):
            batched_graph.other[key] = torch.cat(values)
        else:
            batched_graph.other[key] = torch.stack(values)

    return batched_graph


def is_batch(graph: AtomicGraph) -> bool:
    """
    Does ``graph`` represent a batch of atomic graphs?

    Parameters
    ----------
    graph
        The graph to check.
    """

    return graph.batch is not None


############################### PROPERTIES ###############################


def get_cell_volume(graph: AtomicGraph) -> float:
    """
    Get the volume of the unit cell.
    """

    return torch.det(graph.cell).abs().item()


def number_of_atoms(graph: AtomicGraph) -> int:
    """
    Get the number of atoms in the ``graph``.
    """

    return graph.Z.shape[0]


def number_of_edges(graph: AtomicGraph) -> int:
    """
    Get the number of edges in the ``graph``.
    """

    return graph.neighbour_list.shape[1]


def has_cell(graph: AtomicGraph) -> bool:
    """
    Does ``graph`` represent a structure with a defined unit cell?
    """

    return not torch.allclose(graph.cell, torch.zeros_like(graph.cell))


def neighbour_vectors(graph: AtomicGraph) -> torch.Tensor:
    """
    Get the vector between each pair of atoms specified in the
    ``graph``'s ``"neighbour_list"`` property, respecting periodic
    boundary conditions where present.
    """

    # to simplify the logic below, we'll expand
    # a single graph into a batch of one
    batch: torch.Tensor = torch.zeros_like(graph.Z)
    cell: torch.Tensor = graph.cell.unsqueeze(0)

    # torchscript annoying-ness:
    graph_batch = graph.batch
    if graph_batch is not None:
        cell = graph.cell
        batch = graph_batch

    # avoid tuple de-structuring to keep torchscript happy
    i, j = graph.neighbour_list[0], graph.neighbour_list[1]  # (E,)
    if i.shape[0] == 0:
        return torch.zeros(0, 3, device=graph.R.device)

    cell_per_edge = cell[batch[i]]  # (E, 3, 3)
    distance_offsets = torch.einsum(
        "kl,klm->km",
        graph.neighbour_cell_offsets.to(cell_per_edge.dtype),
        cell_per_edge,
    )  # (E, 3)
    neighbour_positions = graph.R[j] + distance_offsets  # (E, 3)
    return neighbour_positions - graph.R[i]  # (E, 3)


def neighbour_distances(graph: AtomicGraph) -> torch.Tensor:
    """
    Get the distance between each pair of atoms specified in the
    ``graph``'s ``neighbour_list`` property, respecting periodic
    boundary conditions where present.
    """
    return torch.linalg.norm(neighbour_vectors(graph), dim=-1)


def number_of_structures(graph: AtomicGraph) -> int:
    """
    Get the number of structures in the ``graph``.
    """
    # torchscript annoying-ness:
    graph_ptr = graph.ptr
    if graph_ptr is None:
        return 1
    return graph_ptr.shape[0] - 1


def structure_sizes(batch: AtomicGraph) -> torch.Tensor:
    """
    Get the number of atoms in each structure in the ``batch``, of shape
    ``(S,)`` where ``S`` is the number of structures.

    Parameters
    ----------
    batch
        The batch to get the structure sizes for.

    Examples
    --------
    >>> len(graphs)
    3
    >>> [number_of_atoms(g) for g in graphs]
    [3, 4, 5]
    >>> structure_sizes(to_batch(graphs))
    tensor([3, 4, 5])
    """
    # torchscript annoying-ness:
    graph_ptr = batch.ptr
    if graph_ptr is None:
        return torch.scalar_tensor(number_of_atoms(batch))

    return graph_ptr[1:] - graph_ptr[:-1]


def number_of_neighbours(
    graph: AtomicGraph,
    include_central_atom: bool = True,
) -> torch.Tensor:
    """
    Get a tensor, ``T``, of shape ``(N,)``, where ``N`` is the number of atoms
    in the ``graph``, such that ``T[i]`` gives the number of neighbours of atom
    ``i``. If ``include_central_atom`` is ``True``, then the central atom is
    included in the count.

    Parameters
    ----------
    graph
        The graph to get the number of neighbours for.
    include_central_atom
        Whether to include the central atom in the count.
    """

    return sum_over_neighbours(
        torch.ones_like(graph.neighbour_list[0]),
        graph,
    ) + int(include_central_atom)


def available_properties(graph: AtomicGraph) -> List[PropertyKey]:
    """Get the labels that are available on the ``graph``."""
    return [cast(PropertyKey, k) for k in graph.properties]


############################### ACTIONS ###############################


def is_local_property(x: torch.Tensor, graph: AtomicGraph) -> bool:
    """
    Is the property ``x`` local to each atom in the ``graph``?

    Parameters
    ----------
    x
        The property to check.
    graph
        The graph to check the property for.
    """

    return len(x.shape) > 0 and x.shape[0] == number_of_atoms(graph)


def trim_edges(graph: AtomicGraph, cutoff: float) -> AtomicGraph:
    """
    Return a new graph with edges trimmed to be no longer than the ``cutoff``.
    Leaves the original graph unchanged.

    Parameters
    ----------
    graph
        The graph to trim the edges of.
    cutoff
        The maximum distance between atoms to keep the edge.
    """

    existing_cutoff = graph.cutoff
    if existing_cutoff + 1e-5 < cutoff:
        warnings.warn(
            f"Graph already has a cutoff of {existing_cutoff} which is "
            f"less than the requested cutoff of {cutoff}.",
            stacklevel=2,
        )
        return graph
    elif existing_cutoff == cutoff:
        return graph

    distances = neighbour_distances(graph)
    mask = distances <= cutoff
    neighbour_list = graph.neighbour_list[:, mask]
    neighbour_cell_offsets = graph.neighbour_cell_offsets[mask, :]

    # can't use _replace here due to TorchScript
    return AtomicGraph(
        Z=graph.Z,
        R=graph.R,
        cell=graph.cell,
        neighbour_list=neighbour_list,
        neighbour_cell_offsets=neighbour_cell_offsets,
        properties=graph.properties,
        other=graph.other,
        cutoff=cutoff,
        batch=graph.batch,
        ptr=graph.ptr,
    )


def sum_over_central_atom_index(
    p: torch.Tensor,
    central_atom_index: torch.Tensor,
    graph: AtomicGraph,
) -> torch.Tensor:
    r"""
    Efficient, shape-preserving sum of a property, :math:`p`, defined over a
    ``central_atom_index``, to get a per-atom property, :math:`P`,
    such that:

    .. code-block:: python

        # i in central_atom_index
        P[i] == torch.sum(p[central_atom_index == i], dim=0)
        # i not in central_atom_index
        P[i] == torch.zeros_like(p[0])

    .. seealso::

        :func:`sum_over_neighbours` for the explicit case where
        ``p`` is a per-edge property.

    Parameters
    ----------
    p
        The property to sum, of shape ``(Y, ...)``.
    central_atom_index
        The central atoms relevant to each element of ``p``, of shape ``(Y,)``.
    graph
        The graph to sum the property for.

    Returns
    -------
    P: torch.Tensor
        The summed property, of shape ``(N, ...)``.
    """

    N = number_of_atoms(graph)

    # optimised implementations for common cases
    if p.dim() == 1:
        zeros = torch.zeros(N, dtype=p.dtype, device=p.device)
        return zeros.scatter_add(0, central_atom_index, p)

    elif p.dim() == 2:
        C = p.shape[1]
        zeros = torch.zeros(N, C, dtype=p.dtype, device=p.device)
        return zeros.scatter_add(
            0,
            central_atom_index.unsqueeze(1).expand(-1, C),
            p,
        )

    shape = (N,) + p.shape[1:]
    zeros = torch.zeros(shape, dtype=p.dtype, device=p.device)

    if p.shape[0] == 0:
        # return all zeros if there are no atoms
        return zeros

    # create `index`, where index.shape = p.shape
    # and (index[e] == central_atoms[e]).all()
    ones = torch.ones_like(p)
    index = left_aligned_mul(ones, central_atom_index).long()
    return zeros.scatter_add(0, index, p)


def sum_over_neighbours(p: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    r"""
    Shape-preserving sum over neighbours of a per-edge property, :math:`p_{ij}`,
    to get a per-atom property, :math:`P_i`:

    .. math::
        P_i = \sum_{j \in \mathcal{N}_i} p_{ij}

    where:

    * :math:`\mathcal{N}_i` is the set of neighbours of atom :math:`i`.
    * :math:`p_{ij}` is the property of the edge between atoms :math:`i` and
      :math:`j`.
    * :math:`p` is of shape :code:`(E, ...)` and :math:`P` is of shape
      :code:`(N, ...)` where :math:`E` is the number of edges and :math:`N` is
      the number of atoms. :code:`...` denotes any number of additional
      dimensions, including none.
    * :math:`P_i` = 0 if :math:`|\mathcal{N}_i| = 0`.

    Parameters
    ----------
    p
        The per-edge property to sum.
    graph
        The graph to sum the property for.
    """

    return sum_over_central_atom_index(p, graph.neighbour_list[0], graph)


def sum_per_structure(x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    r"""
    Shape-preserving sum of a per-atom property, :math:`p`, to get a
    per-structure property, :math:`P`:

    If a single structure, containing ``N`` atoms, is used, then
    :math:`P = \sum_i p_i`, where:

    * :math:`p_i` is of shape ``(N, ...)``
    * :math:`P` is of shape ``(...)``
    * ``...`` denotes any
      number of additional dimensions, including ``None``.

    If a batch of ``S`` structures, containing a total of ``N`` atoms, is
    used, then :math:`P_k = \sum_{k \in K} p_k`, where:

    * :math:`K` is the collection of all atoms in structure :math:`k`
    * :math:`p_i` is of shape ``(N, ...)``
    * :math:`P` is of shape ``(S, ...)``
    * ``...`` denotes any
      number of additional dimensions, including ``None``.

    Parameters
    ----------
    x
        The per-atom property to sum.
    graph
        The graph to sum the property for.

    Examples
    --------
    Single graph case:

    >>> import torch
    >>> from ase.build import molecule
    >>> from graph_pes.atomic_graph import sum_per_structure, AtomicGraph
    >>> water = molecule("H2O")
    >>> graph = AtomicGraph.from_ase(water, cutoff=1.5)
    >>> # summing over a vector gives a scalar
    >>> sum_per_structure(torch.ones(3), graph)
    tensor(3.)
    >>> # summing over higher order tensors gives a tensor
    >>> sum_per_structure(torch.ones(3, 2, 3), graph).shape
    torch.Size([2, 3])

    Batch case:

    >>> import torch
    >>> from ase.build import molecule
    >>> from graph_pes.atomic_graph import sum_per_structure, AtomicGraph, to_batch
    >>> water = molecule("H2O")
    >>> graph = AtomicGraph.from_ase(water, cutoff=1.5)
    >>> batch = to_batch([graph, graph])
    >>> batch
    AtomicGraphBatch(structures: 2, atoms: 6, edges: 8, has_cell: False)
    >>> # summing over a vector gives a tensor
    >>> sum_per_structure(torch.ones(6), graph)
    tensor([3., 3.])
    >>> # summing over higher order tensors gives a tensor
    >>> sum_per_structure(torch.ones(6, 3, 4), graph).shape
    torch.Size([2, 3, 4])
    """  # noqa: E501

    # torchscript annoying-ness:
    graph_batch = graph.batch
    if graph_batch is not None:
        shape = (number_of_structures(graph),) + x.shape[1:]
        zeros = torch.zeros(shape, dtype=x.dtype, device=x.device)
        return zeros.scatter_add(0, graph_batch, x)
    else:
        return x.sum(dim=0)


def index_over_neighbours(x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    """
    Index a per-atom property, :math:`x`, over the neighbours of each atom in
    the ``graph``.
    """
    return x[graph.neighbour_list[1]]


def divide_per_atom(x: torch.Tensor, graph: AtomicGraph) -> torch.Tensor:
    r"""
    Divide a per-structure property, :math:`X`, by the number of atoms in each
    structure to get a per-atom property, :math:`x`:

    .. math::

        x_i = \frac{X_k}{N_k}

    where:

    * :math:`X` is of shape ``(S, ...)``
    * :math:`x` is of shape ``(N, ...)``
    * :math:`S` is the number of structures
    * :math:`N` is the number of atoms
    """
    return left_aligned_div(x, structure_sizes(graph))



from __future__ import annotations

import warnings

import torch
from torch import nn

from graph_pes.atomic_graph import AtomicGraph
from graph_pes.utils.nn import PerElementParameter
from graph_pes.utils.shift_and_scale import guess_per_element_mean_and_var


class LocalEnergiesScaler(nn.Module):
    """
    Scale the local energies by a per-element scaling factor.

    See :func:`~graph_pes.utils.shift_and_scale.guess_per_element_mean_and_var`
    for how the scaling factors are estimated from the training data.
    """

    def __init__(self):
        super().__init__()
        self.per_element_scaling = PerElementParameter.of_length(
            1,
            default_value=1.0,
            requires_grad=True,
        )
        """
        The per-element scaling factors. 
        (:class:`~graph_pes.utils.nn.PerElementParameter`)
        """

    def forward(
        self,
        local_energies: torch.Tensor,
        graph: AtomicGraph,
    ) -> torch.Tensor:
        """
        Scale the local energies by the per-element scaling factor.
        """
        scales = self.per_element_scaling[graph.Z].squeeze()
        return local_energies.squeeze() * scales

    # add typing for mypy etc
    def __call__(
        self, local_energies: torch.Tensor, graph: AtomicGraph
    ) -> torch.Tensor:
        return super().__call__(local_energies, graph)

    @torch.no_grad()
    def pre_fit(self, graphs: AtomicGraph):
        """
        Pre-fit the per-element scaling factors.

        Parameters
        ----------
        graphs
            The training data.
        """
        if "energy" not in graphs.properties:
            warnings.warn(
                "No energy data found in training data: can't estimate "
                "per-element scaling factors for local energies.",
                stacklevel=2,
            )
            return

        means, variances = guess_per_element_mean_and_var(
            graphs.properties["energy"], graphs
        )
        for Z, var in variances.items():
            self.per_element_scaling[Z] = torch.sqrt(torch.tensor(var))

    def non_decayable_parameters(self) -> list[torch.nn.Parameter]:
        """The ``per_element_scaling`` parameter should not be decayed."""
        return [self.per_element_scaling]

    def __repr__(self):
        return self.per_element_scaling._repr(alias=self.__class__.__name__)



from __future__ import annotations

from itertools import chain
from pathlib import Path
from typing import Callable, Literal

import requests
import torch

from graph_pes import AtomicGraph, GraphPESModel
from graph_pes.atomic_graph import PropertyKey, is_batch, to_batch
from graph_pes.utils.misc import MAX_Z

MACE_KEY_MAPPING: dict[str, PropertyKey] = {
    "node_energy": "local_energies",
    "energy": "energy",
    "forces": "forces",
    "stress": "stress",
    "virials": "virial",
}


class ZToOneHot(torch.nn.Module):
    def __init__(self, elements: list[int]):
        super().__init__()
        self.register_buffer("z_to_index", torch.full((MAX_Z + 1,), -1))
        for i, z in enumerate(elements):
            self.z_to_index[z] = i
        self.num_classes = len(elements)

    def forward(self, Z: torch.Tensor) -> torch.Tensor:
        indices = self.z_to_index[Z]
        return torch.nn.functional.one_hot(indices, self.num_classes)


def _atomic_graph_to_mace_input(
    graph: AtomicGraph,
    z_to_one_hot: Callable[[torch.Tensor], torch.Tensor],
) -> dict[str, torch.Tensor]:
    if not is_batch(graph):
        graph = to_batch([graph])

    assert graph.batch is not None
    assert graph.ptr is not None

    _cell_per_edge = graph.cell[
        graph.batch[graph.neighbour_list[0]]
    ]  # (E, 3, 3)
    _shifts = torch.einsum(
        "kl,klm->km", graph.neighbour_cell_offsets, _cell_per_edge
    )  # (E, 3)
    data = {
        "node_attrs": z_to_one_hot(graph.Z).to(torch.get_default_dtype()),
        "positions": graph.R,
        "cell": graph.cell,
        "edge_index": graph.neighbour_list,
        "unit_shifts": graph.neighbour_cell_offsets,
        "shifts": _shifts,
        "batch": graph.batch,
        "ptr": graph.ptr,
    }
    return {k: v.to(graph.Z.device) for k, v in data.items()}


class MACEWrapper(GraphPESModel):
    """
    Converts any MACE model from the `mace-torch <https://github.com/ACEsuit/mace-torch>`__
    package into a :class:`~graph_pes.GraphPESModel`.

    You can use this to drive MD using LAMMPS, fine-tune MACE models,
    or any functionality that ``graph-pes`` provides.

    Parameters
    ----------
    model
        The MACE model to wrap.

    Examples
    --------
    >>> mace_torch_model = ...  # create your MACE model any-which way
    >>> from graph_pes.interfaces._mace import MACEWrapper
    >>> graph_pes_model = MACEWrapper(mace_torch_model)  # convert to graph-pes
    >>> graph_pes_model.predict_energy(graph)
    torch.Tensor([123.456])
    >>> from graph_pes.utils.calculator import GraphPESCalculator
    >>> calculator = GraphPESCalculator(graph_pes_model)
    >>> calculator.calculate(ase_atoms)

    """

    def __init__(self, model: torch.nn.Module):
        super().__init__(
            model.r_max.item(),
            implemented_properties=[
                "local_energies",
                "energy",
                "forces",
                "stress",
                "virial",
            ],
        )
        self.model = model
        self.z_to_one_hot = ZToOneHot(self.model.atomic_numbers.tolist())

    def forward(self, graph: AtomicGraph) -> dict[PropertyKey, torch.Tensor]:
        return self.predict(
            graph, ["local_energies", "energy", "forces", "stress", "virial"]
        )

    def predict(
        self,
        graph: AtomicGraph,
        properties: list[PropertyKey],
    ) -> dict[PropertyKey, torch.Tensor]:
        raw_predictions = self.model.forward(
            _atomic_graph_to_mace_input(graph, self.z_to_one_hot),
            training=self.training,
            compute_force="forces" in properties,
            compute_stress="stress" in properties,
            compute_virials="virial" in properties,
        )

        predictions: dict[PropertyKey, torch.Tensor] = {
            MACE_KEY_MAPPING[key]: value
            for key, value in raw_predictions.items()
            if key in MACE_KEY_MAPPING
        }
        if not is_batch(graph):
            for p in ["energy", "stress", "virial"]:
                if p in properties:
                    predictions[p] = predictions[p].squeeze()
        return {k: v for k, v in predictions.items() if k in properties}


def _fix_dtype(model: torch.nn.Module, dtype: torch.dtype) -> None:
    for tensor in chain(
        model.parameters(),
        model.buffers(),
    ):
        if tensor.dtype.is_floating_point:
            tensor.data = tensor.data.to(dtype)


def _get_dtype(
    precision: Literal["float32", "float64"] | None,
) -> torch.dtype:
    if precision is None:
        return torch.get_default_dtype()
    return {"float32": torch.float32, "float64": torch.float64}[precision]


def mace_mp(
    model: Literal["small", "medium", "large"],
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    """
    Donwload a MACE-MP model and convert it for use with ``graph-pes``.

    Internally, we use the `foundation_models <https://mace-docs.readthedocs.io/en/latest/guide/foundation_models.html>`__
    functionality from the `mace-torch <https://github.com/ACEsuit/mace-torch>`__ package.

    Please cite the following if you use this model:

    - MACE-MP by Ilyes Batatia, Philipp Benner, Yuan Chiang, Alin M. Elena,
      Dávid P. Kovács, Janosh Riebesell, et al., 2023, arXiv:2401.00096
    - MACE-Universal by Yuan Chiang, 2023, Hugging Face, Revision e5ebd9b,
      DOI: 10.57967/hf/1202, URL: https://huggingface.co/cyrusyc/mace-universal
    - Matbench Discovery by Janosh Riebesell, Rhys EA Goodall, Philipp Benner,
      Yuan Chiang, Alpha A Lee, Anubhav Jain, Kristin A Persson, 2023,
      arXiv:2308.14920

    Parameters
    ----------
    model
        The size of the MACE-MP model to download.
    precision
        The precision of the model. If ``None``, the default precision
        of torch will be used (you can set this when using ``graph-pes-train``
        via ``general/torch/dtype``)
    """  # noqa: E501
    from mace.calculators.foundations_models import mace_mp

    dtype = _get_dtype(precision)
    precision_str = {torch.float32: "float32", torch.float64: "float64"}[dtype]

    mace_torch_model = mace_mp(
        model,
        device="cpu",
        default_dtype=precision_str,
        return_raw_model=True,
    )
    assert isinstance(mace_torch_model, torch.nn.Module)
    _fix_dtype(mace_torch_model, dtype)
    return MACEWrapper(mace_torch_model)


def mace_off(
    model: Literal["small", "medium", "large"],
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    """
    Download a MACE-OFF model and convert it for use with ``graph-pes``.

    If you use this model, please cite the relevant paper by Kovacs et.al., arXiv:2312.15211

    Parameters
    ----------
    model
        The size of the MACE-OFF model to download.
    precision
        The precision of the model.
    """  # noqa: E501
    from mace.calculators.foundations_models import mace_off

    dtype = _get_dtype(precision)
    precision_str = {torch.float32: "float32", torch.float64: "float64"}[dtype]

    mace_torch_model = mace_off(
        model,
        device="cpu",
        default_dtype=precision_str,
        return_raw_model=True,
    )
    assert isinstance(mace_torch_model, torch.nn.Module)
    _fix_dtype(mace_torch_model, dtype)
    # un freeze all parameters
    for p in mace_torch_model.parameters():
        p.requires_grad = True
    return MACEWrapper(mace_torch_model)


def go_mace_23(
    precision: Literal["float32", "float64"] | None = None,
) -> MACEWrapper:
    """
    Download the `GO-MACE-23 model <https://doi.org/10.1002/anie.202410088>`__
    and convert it for use with ``graph-pes``.

    .. note::

        This model is only for use on structures containing Carbon, Hydrogen and
        Oxygen. Attempting to use on structures with other elements will raise
        an error.

    If you use this model, please cite the following:

    .. code-block:: bibtex

        @article{El-Machachi-24,
            title = {Accelerated {{First-Principles Exploration}} of {{Structure}} and {{Reactivity}} in {{Graphene Oxide}}},
            author = {{El-Machachi}, Zakariya and Frantzov, Damyan and Nijamudheen, A. and Zarrouk, Tigany and Caro, Miguel A. and Deringer, Volker L.},
            year = {2024},
            journal = {Angewandte Chemie International Edition},
            volume = {63},
            number = {52},
            pages = {e202410088},
            doi = {10.1002/anie.202410088},
        }

    """  # noqa: E501

    dtype = _get_dtype(precision)

    url = "https://github.com/zakmachachi/GO-MACE-23/raw/refs/heads/main/models/fitting/potential/iter-12-final-model/go-mace-23.pt"
    save_path = Path.home() / ".graph-pes" / "go-mace-23.pt"
    save_path.parent.mkdir(parents=True, exist_ok=True)

    if not save_path.exists():
        print(f"Downloading GO-MACE-23 model to {save_path}")
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for bad status codes

        with open(save_path, "wb") as file:
            file.write(response.content)

    print(f"Loading GO-MACE-23 model from {save_path}")
    mace_torch_model = torch.load(
        save_path, weights_only=False, map_location=torch.device("cpu")
    )
    for p in mace_torch_model.parameters():
        p.data = p.data.to(dtype)
    model = MACEWrapper(mace_torch_model)

    return model



from __future__ import annotations

from pytorch_lightning.loggers import Logger as PTLLogger

from graph_pes.atomic_graph import (
    AtomicGraph,
    number_of_atoms,
    number_of_structures,
)
from graph_pes.graph_pes_model import GraphPESModel
from graph_pes.models.addition import AdditionModel
from graph_pes.utils.logger import logger
from graph_pes.utils.nn import learnable_parameters


def log_model_info(
    model: GraphPESModel,
    ptl_logger: PTLLogger | None = None,
) -> None:
    """Log the number of parameters in a model."""

    logger.debug(f"Model:\n{model}")

    if isinstance(model, AdditionModel):
        model_names = [
            f"{given_name} ({component.__class__.__name__})"
            for given_name, component in model.models.items()
        ]
        params = [
            learnable_parameters(component)
            for component in model.models.values()
        ]
        width = max(len(name) for name in model_names)
        info_str = "Number of learnable params:"
        for name, param in zip(model_names, params):
            info_str += f"\n    {name:<{width}}: {param:,}"
        logger.info(info_str)

    else:
        logger.info(
            f"Number of learnable params : {learnable_parameters(model):,}"
        )

    if ptl_logger is not None:
        all_params = sum(p.numel() for p in model.parameters())
        learnable_params = learnable_parameters(model)
        ptl_logger.log_metrics(
            {
                "n_parameters": all_params,
                "n_learnable_parameters": learnable_params,
            }
        )


def sanity_check(model: GraphPESModel, batch: AtomicGraph) -> None:
    outputs = model.get_all_PES_predictions(batch)

    N = number_of_atoms(batch)
    S = number_of_structures(batch)
    expected_shapes = {
        "local_energies": (N,),
        "forces": (N, 3),
        "energy": (S,),
        "stress": (S, 3, 3),
        "virial": (S, 3, 3),
    }

    incorrect = []
    for key, value in outputs.items():
        if value.shape != expected_shapes[key]:
            incorrect.append((key, value.shape, expected_shapes[key]))

    if len(incorrect) > 0:
        raise ValueError(
            "Sanity check failed for the following outputs:\n"
            + "\n".join(
                f"{key}: {value} != {expected}"
                for key, value, expected in incorrect
            )
        )

    if batch.cutoff < model.cutoff:
        logger.error(
            "Sanity check failed: you appear to be training on data "
            f"composed of graphs with a cutoff ({batch.cutoff}) that is "
            f"smaller than the cutoff used in the model ({model.cutoff}). "
            "This is almost certainly not what you want to do?",
        )


VALIDATION_LOSS_KEY = "valid/loss/total"
